#include "mublas_devel.h"

/* ***********************************************
	Default Recipe
   *********************************************** */   

void
mublasRecipeDefault (mublasHandle_t *mh) {
	// --------------------------------------------------------
	mh->WrpOcpSuf = 0.0;//14.0 * (float)mh->WrpGran / (float)mh->WrpPerMpPmax; 
	mh->BlkOcpSuf = 0.0 / (float)mh->BlkPerMpPmax;
	mh->GrdOcpSuf = 1.0;
	mh->TyPerTxMax = 1.0; //1024=none
	// --------------------------------------------------------
	mh->TxMin = 0; //0=auto:=TxStep
	mh->TxMax = 0;
	mh->TxStep = 0; //0=auto:128/sizeof(val)
	mh->ThMin = 0; //0=auto:ThMin=TxStep
	mh->ThMax = 0; //0=auto:maxThreadsPerBlock
	mh->ThStep = 0; //0=auto=TxStep
	// --------------------------------------------------------
}

// compute shmem availability
__host__ __inline__ void getShmSize (mublasHandle_t* mh) {
	size_t res;
	const size_t preferS = mh->devprop.sharedMemPerMultiprocessor;
    const size_t preferL = mh->devprop.sharedMemPerMultiprocessor - 32768;
    const size_t preferE = (preferS + preferL) / 2;
    switch (mh->devprop.major) {
		case 1:
			mh->ShmPerBlkMax = preferS; break;
		case 2:
			switch (mh->func_cache) {
				case cudaFuncCachePreferNone:   mh->ShmPerBlkMax = preferS; break;
				case cudaFuncCachePreferShared: mh->ShmPerBlkMax = preferS; break;
				case cudaFuncCachePreferL1:		mh->ShmPerBlkMax = preferL; break;
				case cudaFuncCachePreferEqual:	mh->ShmPerBlkMax = preferS; break;
				default:						mh->ShmPerBlkMax = preferS; break;
			}
			break;
		case 3:
			switch (mh->func_cache) {
				case cudaFuncCachePreferNone:   mh->ShmPerBlkMax = preferS; break;
				case cudaFuncCachePreferShared: mh->ShmPerBlkMax = preferS; break;
				case cudaFuncCachePreferL1:		mh->ShmPerBlkMax = preferL; break;
				case cudaFuncCachePreferEqual:	mh->ShmPerBlkMax = preferE; break;
				default:						mh->ShmPerBlkMax = preferS; break;
			}
			break;
		case 5:
			mh->ShmPerBlkMax = preferS; break;
		default: 
			mh->ShmPerBlkMax = preferS; break;
	}
}

__host__ __inline__ size_t ThSizeCheck (mublasHandle_t *mh) {
	if ((mh->Tx * mh->Ty) % 32 == 0) return 1;
	return 0;
}

/* ***********************************************
	Occupancy Calculation
   *********************************************** */   

__host__ __inline__ void getOccupancies (mublasHandle_t *mh) {
	const size_t RegPerWrp = ceilingEx ((float)(mh->funcatb.numRegs * WrpSize), mh->RegUnit); // Usage of registers per warp
	const size_t WrpPerBlk = ceil ((float)(mh->Tx * mh->Ty) / (float)WrpSize); // Warps per block
	const size_t BlkPerMpRmax = floor (min (floorEx ((float)mh->devprop.regsPerMultiprocessor / (float)RegPerWrp, mh->WrpGran), mh->WrpPerMpPmax) / (float)WrpPerBlk);
	const size_t BlkPerMpSmax = (mh->ShmPerBlk>0) ? floor ((float)mh->ShmPerBlkMax / (float)mh->ShmPerBlk) : mh->BlkPerMpPmax;
	mh->BlkPerMpMax = min (min (BlkPerMpRmax, BlkPerMpSmax), mh->BlkPerMpPmax);
	mh->WrpOcp = (float)(mh->BlkPerMpMax * WrpPerBlk) / (float)mh->WrpPerMpPmax; 
	mh->BlkOcp = (float)(mh->BlkPerMpMax) / (float)mh->BlkPerMpPmax;
	mh->GrdOcp = (float)(mh->Bx * mh->By) / (float)ceilingEx (mh->Bx * mh->By, mh->BlkPerMpMax * mh->devprop.multiProcessorCount);
}

__host__ void determine_best_thread_block_size_2d (mublasHandle_t *mh) {
	size_t TxTmp, TyTmp, TxStart, TxStop, TxStep;
	float GrdOcpPrev = 0., BlkOcpPrev = 0., WrpOcpPrev = 0.;
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;

	TxTmp = mh->TxMax;
	TyTmp = ceil ((float)mh->ThMin / TxTmp);
	mh->count = 0;
	if (mh->autotuning) printf ("%d,",mh->n);
	// --- Tx ------------------------------------------------
	mh->Tx = mh->TxMax;
	while (mh->Tx >= mh->TxMin) {
		// --- Ty -----------------------------------------------
		mh->Ty = min (max (1, floor ((float)mh->ThMax / (float)mh->Tx)), mh->Tx * mh->TyPerTxMax);
		while (mh->Ty >= ceil ((float)mh->ThMin / (float)mh->Tx)) {
			if (ThSizeCheck(mh)) {
				mh->count++;
				getBlkPerGrd (mh);
				getShmPerBlk (mh);
				getOccupancies (mh);
				if (mh->autotuning) printf ("%dx%d,", mh->Tx, mh->Ty);
				if (mh->WrpOcp >= min(WrpOcpPrev,mh->WrpOcpSuf) &&
					mh->BlkOcp >= min(BlkOcpPrev,mh->BlkOcpSuf) &&
					mh->GrdOcp >= min(GrdOcpPrev,mh->GrdOcpSuf)) {
					GrdOcpPrev = mh->GrdOcp;
					WrpOcpPrev = mh->WrpOcp;
					BlkOcpPrev = mh->BlkOcp;
					TxTmp = mh->Tx;
					TyTmp = mh->Ty;
				}
			}
			mh->Ty -= 1;
		}
		mh->Tx -= mh->TxStep;
	}
	mh->Tx = TxTmp;
	mh->Ty = TyTmp;
	if (mh->count == 0)
		if (mh->autotuning) printf ("%dx%d,", mh->Tx, mh->Ty);
}

__host__ void determine_best_thread_block_size_1d (mublasHandle_t *mh) {
	size_t TxTmp;
	float GrdOcpPrev = 0., BlkOcpPrev = 0., WrpOcpPrev = 0.;
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;

	TxTmp = mh->TxMax;
	mh->count = 0;
	mh->Ty = 1;
	if (mh->autotuning) printf ("%d,",mh->n);
	// --- Tx ------------------------------------------------
	mh->Tx = mh->TxMax;
	while (mh->Tx >= mh->TxMin) {
		if (ThSizeCheck(mh)) {
			mh->count++;
			getBlkPerGrd (mh);
			getShmPerBlk (mh);
			getOccupancies (mh);
			if (mh->autotuning) printf ("%dx%d,", mh->Tx, mh->Ty);
			if (mh->WrpOcp >= min(WrpOcpPrev,mh->WrpOcpSuf) &&
				mh->BlkOcp >= min(BlkOcpPrev,mh->BlkOcpSuf) &&
				mh->GrdOcp >= min(GrdOcpPrev,mh->GrdOcpSuf)) {
				GrdOcpPrev = mh->GrdOcp;
				BlkOcpPrev = mh->BlkOcp;
				WrpOcpPrev = mh->WrpOcp;
				TxTmp = mh->Tx;
			}
		}
		mh->Tx -= mh->TxStep;
	}
	mh->Tx = TxTmp;
	mh->Ty = 1;
	if (mh->count == 0)
		if (mh->autotuning) printf ("%dx%d,", mh->Tx, mh->Ty);
}

__host__ void determine_best_thread_block_size_nvidia (mublasHandle_t *mh) {
	cudaOccError err_t; 
	cudaOccDeviceProp occdevprop = (cudaOccDeviceProp) mh->devprop;
	cudaOccFuncAttributes occfuncatb = mh->funcatb;
	cudaOccDeviceState occdevstat;
	occdevstat.cacheConfig = (cudaOccCacheConfig) mh->func_cache;
	void (*getShmPerBlk)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getShmPerBlk (mh);
	int minGridSize, blockSize;

	err_t = cudaOccMaxPotentialOccupancyBlockSize(
		&minGridSize, &blockSize, // out
		&occdevprop, &occfuncatb, &occdevstat, mh->ShmPerBlk);
	printf ("minGridSize = %d, blockSize = %d, err_t = %d\n", minGridSize, blockSize, err_t);
}

// phase 2: determining optimal thread-block size
__host__ void mublas_optimizer (mublasHandle_t *mh) {
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;
	getShmSize(mh);

	// --- Initial Tx & Ty -------------------------------
	// 1. TxStep (0=auto)
	if (mh->TxStep == 0) mh->TxStep = 128 / (mh->N_one * mh->N_T); // 128bytes/16bytes(vec)=8thread
	// 2. ThStep (0=auto)
	if (mh->ThStep == 0) mh->ThStep = mh->TxStep;//WrpSize;
	// 3. ThMin (0=auto)
	if (mh->ThMin == 0) mh->ThMin = mh->WrpGran * WrpSize;
	// 4. ThMax
	if (mh->ThMax == 0) mh->ThMax = mh->funcatb.maxThreadsPerBlock;
	// 5. TxMin
	if (mh->TxMin == 0) mh->TxMin = mh->TxStep;
	if (mh->BlkDim == 1) mh->TxMin = max (mh->TxMin, mh->ThMin);
	// 6. TxMax
	if (mh->TxMax == 0) mh->TxMax = mh->ThMax;
	// ---------------------------------------------------

	if (mh->TxStatic == 0) { 
		if      (mh->BlkDim == 2) determine_best_thread_block_size_2d (mh);
		else if (mh->BlkDim == 1) determine_best_thread_block_size_1d (mh);
		else printf ("mublas error: unsupported BlkDim\n");
	} else { // Kernel is launched with specified Tx and Ty (static mode)
		mh->Tx = mh->TxStatic;
		mh->Ty = mh->TyStatic;
		if (mh->Tx * mh->Ty > mh->funcatb.maxThreadsPerBlock) {
			printf ("mublas error: too many resource requested\n");
			exit (1);
		}
	}

//	determine_best_thread_block_size_nvidia (mh); // debug, cuda_occupancy.h

	getBlkPerGrd (mh);
	getShmPerBlk (mh);
	getOccupancies (mh);
}

/* ***********************************************
	mublas functions
   *********************************************** */   

void
mublasCreate (mublasHandle_t *mh) {
    int32_t driverVersion = 0;
	int32_t runtimeVersion = 0;
	int32_t devid;
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);

	// Device Properties
	cudaDeviceProp devprop;  
	cudaGetDevice (&devid);
	cudaGetDeviceProperties (&devprop, devid);
	mh->devprop = devprop;
	mh->cc = devprop.major * 100 + devprop.minor * 10;
	mh->driverVersion = driverVersion;
	mh->runtimeVersion = runtimeVersion;
	mh->streamId = 0;

	// work space allocation
	float *w;
	cudaMalloc ((void **) &w, sqrt((float)devprop.totalGlobalMem/(sizeof(double)*4))*sizeof(double)*4);
	mh->work_addr = w;
	float *w2;
	cudaMalloc ((void **) &w2, sqrt((float)devprop.totalGlobalMem/(sizeof(double)*4))*sizeof(double)*4);
	mh->work_addr2 = w2;

	// --------------------------------------------------------
	// init
	mh->mublasVersion = MUBLAS_VERSION;
	mh->mublaserror = 1111; // set 1111 when mublasCreate called
	mh->kernelnumber = 0;
	mh->count = 0;
	mh->debug = 0;
	mh->autotuning = 0; // mode=t
	mh->WrpOcp = 0.;
	mh->BlkOcp = 0.;
	mh->GrdOcp = 0.;
	mh->Tx = 0;
	mh->Ty = 0;
	mh->Bx = 0;
	mh->By = 0;
	// --------------------------------------------------------

	// default value (architecture dependent)
	switch (mh->cc) {
		case 200:
			mh->WrpPerMpPmax = 48;
			mh->BlkPerMpPmax = 8;
			mh->RegUnit = 64;
			mh->ShmUnit = 128;
			mh->WrpGran = 2;
			mh->Cores = 32;
			mh->DPCores = 16;
			break;
		case 210:
			mh->WrpPerMpPmax = 48;
			mh->BlkPerMpPmax = 8;
			mh->RegUnit = 64;
			mh->ShmUnit = 128;
			mh->WrpGran = 2;
			mh->Cores = 48;
			mh->DPCores = 6;
			break;
		case 300:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 320:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 350:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 370:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 500:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 32;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 128;
			mh->DPCores = 4;
			break;
		case 520:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 32;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 128;
			mh->DPCores = 4;
			break;
		default: // non-supported architecture
			mh->mublaserror = 9999;
			break;
	}
	mublasRecipeDefault (mh);
}

void
mublasSetStream (mublasHandle_t *mh, cudaStream_t stream) {
	mh->streamId = stream;
}

void
mublasDestroy (mublasHandle_t *mh) {
	cudaFree (mh->work_addr);
	cudaFree (mh->work_addr2);
}

/* ***********************************************
	mublas functions (multi-GPU support)
   *********************************************** */   

void mublasSetMatrixMgpu1D (size_t m, size_t n, void* hstAddr, size_t ld_hst, void** devAddr, size_t ld_dev, uint32_t ngpus, size_t bs, size_t word_size) {
	uint32_t igpu;
	char* devAddr_;
	char* hstAddr_ = (char*)hstAddr;
	size_t nr, hst_offset, dev_offset;
	dev_offset = 0;
	hst_offset = 0;

	nr = n;
	igpu = 0;
	while (nr > bs) {
		cudaSetDevice (igpu%ngpus);
		devAddr_ = (char*)devAddr[igpu%ngpus];
		cublasSetMatrixAsync (m, bs, word_size, hstAddr_+hst_offset, ld_hst, devAddr_+dev_offset, ld_dev, 0);
		// update
		nr -= bs;
		igpu++;
		hst_offset += ld_hst * bs * word_size;
		if (igpu%ngpus == 0) dev_offset += ld_dev * bs * word_size;
	} 
	cudaSetDevice (igpu%ngpus);
	devAddr_ = (char*)devAddr[igpu%ngpus];
	cublasSetMatrixAsync (m, nr, word_size, hstAddr_+hst_offset, ld_hst, devAddr_+dev_offset, ld_dev, 0);

	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize ();
	}
}

void mublasSetVectorMgpu1D (size_t n, void* hstAddr, void** devAddr, uint32_t ngpus, size_t bs, size_t word_size) {
	uint32_t igpu;
	char* devAddr_;
	char* hstAddr_ = (char*)hstAddr;
	size_t nr, hst_offset, dev_offset;
	dev_offset = 0;
	hst_offset = 0;

	nr = n;
	igpu = 0;
	while (nr > bs) {
		cudaSetDevice (igpu%ngpus);
		devAddr_ = (char*)devAddr[igpu%ngpus];
		cublasSetVectorAsync (bs, word_size, hstAddr_+hst_offset, 1, devAddr_+dev_offset, 1, 0);
		// update
		nr -= bs;
		igpu++;
		hst_offset += bs * word_size;
		if (igpu%ngpus == 0) dev_offset += bs * word_size;
	} 
	cudaSetDevice (igpu%ngpus);
	devAddr_ = (char*)devAddr[igpu%ngpus];
	cublasSetVectorAsync (nr, word_size, hstAddr_+hst_offset, 1, devAddr_+dev_offset, 1, 0);

	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize ();
	}
}

void mublasGetVectorMgpu1D (size_t n, void** devAddr, void* hstAddr, uint32_t ngpus, size_t bs, size_t word_size) {
	uint32_t igpu;
	char* devAddr_;
	char* hstAddr_ = (char*)hstAddr;
	size_t nr, hst_offset, dev_offset;
	dev_offset = 0;
	hst_offset = 0;

	nr = n;
	igpu = 0;
	while (nr > bs) {
		cudaSetDevice (igpu%ngpus);
		devAddr_ = (char*)devAddr[igpu%ngpus];
		cublasGetVectorAsync (bs, word_size, devAddr_+dev_offset, 1, hstAddr_+hst_offset, 1, 0);
		// update
		nr -= bs;
		igpu++;
		hst_offset += bs * word_size;
		if (igpu%ngpus == 0) dev_offset += bs * word_size;
	} 
	cudaSetDevice (igpu%ngpus);
	devAddr_ = (char*)devAddr[igpu%ngpus];
	cublasGetVectorAsync (nr, word_size, devAddr_+dev_offset, 1, hstAddr_+hst_offset, 1, 0);

	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize ();
	}
}

void mublasMallocMatrixMgpu1D (size_t m, size_t n, void** devAddr, size_t ld_dev, uint32_t ngpus, size_t bs, size_t word_size) {
	uint32_t igpu;
	size_t pitch, nr, ni[NGPU_MAX] = {0};

	nr = n;
	igpu = ngpus;
	while (nr >= bs) {
		ni[igpu%ngpus] += bs;
		nr -= bs;
		igpu++;
	} 
	ni[igpu%ngpus] += nr;

	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaMalloc ((void **) &devAddr[igpu], word_size * ld_dev * ni[igpu]);
	}
	
	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize ();
	}
}

/* ***********************************************
	mublas functions (multi-GPU support, 1D-block)
   *********************************************** */   

void mublasMallocVectorMgpu (mublasMgpuVectorHandle_t* mmah, size_t n, size_t word_size, size_t ngpus) {
	size_t i;
	mmah->word_size = word_size;
	mmah->ngpus = ngpus;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		mmah->n[i] = n;
		cudaMalloc ((void **) &mmah->devAddr[i], mmah->word_size * mmah->n[i]);
	}
}

void mublasMallocVectorMgpuDistributed (mublasMgpuVectorHandle_t* mmah, size_t n, size_t word_size, size_t ngpus) {
	size_t i;
	mmah->word_size = word_size;
	mmah->ngpus = ngpus;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		mmah->n[i] = (i == ngpus-1) ? n-ceil(n/ngpus)*(ngpus-1) : ceil(n/ngpus);
		cudaMalloc ((void **) &mmah->devAddr[i], mmah->word_size * mmah->n[i]);
	}
}

void mublasMallocMatrixMgpuDistributed (mublasMgpuMatrixHandle_t* mmah, char trans, size_t m, size_t n, size_t word_size, size_t ngpus) {
	size_t i;
	size_t pitch;
	mmah->word_size = word_size;
	mmah->ngpus = ngpus;
	if (trans == 'N' || trans == 'n') {
		for (i = 0; i < mmah->ngpus; i++) {
			mmah->m[i] = (i == ngpus-1) ? m-ceil(m/ngpus)*(ngpus-1) : ceil(m/ngpus);
			mmah->n[i] = n;
			cudaSetDevice (i);
			cudaMallocPitch ((void **) &mmah->devAddr[i], &pitch, mmah->word_size * mmah->m[i], mmah->n[i]);
			mmah->ld[i] = pitch/mmah->word_size;
		}
	} else {
		for (i = 0; i < mmah->ngpus; i++) {
			mmah->n[i] = (i == ngpus-1) ? n-ceil(n/ngpus)*(ngpus-1) : ceil(n/ngpus);
			mmah->m[i] = m;
			cudaSetDevice (i);
			cudaMallocPitch ((void **) &mmah->devAddr[i], &pitch, mmah->word_size * mmah->m[i], mmah->n[i]);
			mmah->ld[i] = pitch/mmah->word_size;
		}
	}
}

void mublasSetVectorMgpu (mublasMgpuVectorHandle_t* mmah, void* hstAddr) {
	size_t i;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cublasSetVector (mmah->n[i], mmah->word_size, hstAddr, 1, mmah->devAddr[i], 1);
	}
}

void mublasSetVectorMgpuDistributed (mublasMgpuVectorHandle_t* mmah, void* hstAddr) {
	size_t i;
	char* mgpu_hstAddr = (char*)hstAddr;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cublasSetVector (mmah->n[i], mmah->word_size, mgpu_hstAddr, 1, mmah->devAddr[i], 1);
		mgpu_hstAddr += mmah->n[i] * mmah->word_size;
	}
}

void mublasSetMatrixMgpuDistributed (mublasMgpuMatrixHandle_t* mmah, char trans, void* hstAddr, size_t ld_hst) {
	size_t i;
	char* mgpu_hstAddr = (char*)hstAddr;
	if (trans == 'N' || trans == 'n') {
		for (i = 0; i < mmah->ngpus; i++) {
			cudaSetDevice (i);
			cublasSetMatrix (mmah->m[i], mmah->n[i], mmah->word_size, mgpu_hstAddr, ld_hst, mmah->devAddr[i], mmah->ld[i]);
			mgpu_hstAddr += mmah->m[i] * mmah->word_size;
		}
	} else {
		for (i = 0; i < mmah->ngpus; i++) {
			cudaSetDevice (i);
			cublasSetMatrix (mmah->m[i], mmah->n[i], mmah->word_size, mgpu_hstAddr, ld_hst, mmah->devAddr[i], mmah->ld[i]);
			mgpu_hstAddr += ld_hst * mmah->n[i] * mmah->word_size;
		}
	}
}

void mublasGetVectorMgpuDistributed (mublasMgpuVectorHandle_t* mmah, void* hstAddr) {
	size_t i;
	char* mgpu_hstAddr = (char*)hstAddr;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cublasGetVector (mmah->n[i], mmah->word_size, mmah->devAddr[i], 1, mgpu_hstAddr, 1);
		mgpu_hstAddr += mmah->n[i] * mmah->word_size;
	}
}

void mublasFreeMatrixMgpu (mublasMgpuMatrixHandle_t* mmah) {
	size_t i;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cudaFree (mmah->devAddr[i]);
	}
}

void mublasFreeVectorMgpu (mublasMgpuVectorHandle_t* mmah) {
	size_t i;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cudaFree (mmah->devAddr[i]);
	}
}

