#ifndef MUBLAS_DEVEL_H
#define MUBLAS_DEVEL_H

// ***** MUBLAS Version **************************
#define MUBLAS_VERSION						105024
// ***********************************************

// ***** size_t / int32_t ************************
// "SIZE_T": integer type on GPU kernels
// we should not use uint32_t on CUDA kernels??
#define SIZE_T								int32_t 
// ***********************************************

// ***** USE_INLINE_ASM **************************
#define USE_INLINE_ASM						1
// !!! DO NOT USE with nvcc <= 6.5 
// ***********************************************

#include <stdio.h>
#include <stdlib.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <cuda_runtime_api.h>
#include <cublas_v2.h>
#include <cuda_occupancy.h>

// Threads per warp (warp size)
#define WrpSize 32

#define min(a, b) ((a) < (b) ? (a) : (b))
#define max(a, b) ((a) > (b) ? (a) : (b))

#include "../include/mublas.h"

// Ceiling and floor function (Excel compatible) -----------------------
__host__ __inline__ size_t ceilingEx (const float n, const size_t c) {
	return (size_t)ceil(n/c)*c;
}

__host__ __inline__ size_t floorEx (const float n, const size_t c) {
	return (size_t)floor(n/c)*c;
}
// ---------------------------------------------------------------------

// vector address calculation for mgpu -------------------------------
__device__ __forceinline__ SIZE_T xaddr (SIZE_T addr, SIZE_T ngpus, SIZE_T gpuid, SIZE_T bs) {
	return (bs * gpuid) + (addr / bs) * (bs * ngpus) + (addr % bs);
}
// ---------------------------------------------------------------------

extern __host__ void mublasRecipeDefault (mublasHandle_t* mh);
extern __host__ void getOccupancies ( mublasHandle_t* mh);
extern __host__ void mublas_optimizer (mublasHandle_t* mh);
extern __host__ void getShmSize ( mublasHandle_t* mh);

#ifdef PREC_S
#include "prec_s.h"
#endif
#ifdef PREC_D
#include "prec_d.h"
#endif
#ifdef PREC_C
#include "prec_c.h"
#endif
#ifdef PREC_Z
#include "prec_z.h"
#endif
#ifdef PREC_W
#include "prec_w.h"
#endif
#ifdef PREC_U
#include "prec_u.h"
#endif
#ifdef PREC_B
#include "prec_b.h"
#endif
#ifdef PREC_V
#include "prec_v.h"
#endif
#ifdef PREC_DB
#include "prec_db.h"
#endif

#endif
