/* ***********************************************

DOT

  *********************************************** */   

#include "mublas_devel.h"
#include "mublas_dot_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   
__host__ int32_t
MUBLAS_DOT (
	mublasHandle_t* mh,	
	const size_t n,
	const TYPE0 *x,
	const size_t incx,
	const TYPE0 *y,
	const size_t incy,
	TYPE0 *result
) {
	uint32_t kernel_number = 0;
	void (*kernel)  (const SIZE_T, const TYPE0*, const TYPE0*, TYPE0*);
	void (*kernelb) (const SIZE_T, TYPE0*);
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	if (n == 0) 
		return 0;
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration 
	kernel_number = 999;
	mh->n = n;
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if ((((long)x%(128/sizeof(TYPE0)))%16 != 0) || 
		(((long)y%(128/sizeof(TYPE0)))%16 != 0)) mh->N_T = 1;
	if (incx == 1 && incy == 1) 
		kernel_number = 100 + mh->N_T * 10;
	// debug
	if (mh->kerneldebug != 0) {
		kernel_number = mh->kerneldebug;
		switch (kernel_number) {
			case 110: mh->N_T = 1; break;
			case 120: mh->N_T = 2; break;
			case 140: mh->N_T = 4; break;
		}
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Unsupported operations (CUBLAS)
	if (kernel_number == 999) {
		#if defined (CUBLAS_DOT)
		cublasHandle_t ch;
		cublasCreate (&ch);
		cublasSetStream (ch, mh->streamId);
		cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
		CUBLAS_DOT (ch, n, x, incx, y, incy, result);
		#else
		printf ("mublas error: not implemented");
		#endif
		mh->kernelnumber = kernel_number;
		return 0;
	}
	// ---------------------------------------------
	
	// ---------------------------------------------
	// 4. Kernel Launch Configuration
	if (incx == 1 && incy == 1) {
		mh->BlkDim = 1;
		mh->func_getShmPerBlk = &getShmPerBlk_DOT;
		mh->func_getBlkPerGrd = &getBlkPerGrd_DOT;
		mh->func_cache = cudaFuncCachePreferShared;
		#include "mublas_dot.incl"
	}
	kernelb = &MUBLAS_SUM_DEVICE_s1;
	cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_SUM_DEVICE_s1);
	cudaFuncSetCacheConfig (MUBLAS_SUM_DEVICE_s1, mh->func_cache);
	switch (kernel_number) {
		case 110:
			kernel = &MUBLAS_DOT_DEVICE_s1;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_DOT_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_DOT_DEVICE_s1, mh->func_cache);
			break;
		case 120:
			kernel = &MUBLAS_DOT_DEVICE_v2;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_DOT_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_DOT_DEVICE_v2, mh->func_cache);
			//kernelb = &MUBLAS_SUM_DEVICE_v2;
			//cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_SUM_DEVICE_v2);
			//cudaFuncSetCacheConfig (MUBLAS_SUM_DEVICE_v2, mh->func_cache);
			break;
		case 140:
			kernel = &MUBLAS_DOT_DEVICE_v4;
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_DOT_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_DOT_DEVICE_v4, mh->func_cache);
			//kernelb = &MUBLAS_SUM_DEVICE_v4;
			//cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_SUM_DEVICE_v4);
			//cudaFuncSetCacheConfig (MUBLAS_SUM_DEVICE_v4, mh->func_cache);
			break;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 5. Thread-Block Size Adjust and Kernel Launch
	mublas_optimizer (mh);
	dim3 threads = dim3 (mh->Tx, mh->Ty);
	dim3 grid = dim3 (mh->Bx, mh->By);
	if (mh->autotuning!=1) kernel <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (n, x, y, (TYPE0*)mh->work_addr);
	size_t _Tx = mh->Tx;
	size_t _Bx = mh->Bx;
	// SUM
	//mh->Tx = 128;
	mh->N_T = 1;
	mh->n = mh->Bx;
	while (mh->n > 1){
		mh->func_getShmPerBlk (mh);
		mh->func_getBlkPerGrd (mh);
		mh->func_getTxSuf (mh);
		threads = dim3 (mh->Tx);
		grid = dim3 (mh->Bx);
		kernelb <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (mh->n, (TYPE0*)mh->work_addr);
		mh->n = mh->Bx;
	}
	cudaMemcpy (result, (TYPE0*)mh->work_addr, sizeof(TYPE0), cudaMemcpyDeviceToHost);
	mh->Tx = _Tx;
	mh->Bx = _Bx;
	// ---------------------------------------------

	mh->kernelnumber = kernel_number;
	return 0;
}

