/* ***********************************************

DOT

memo:
Actually v2&v4 kernels are not needed.
v1 kernel with AT is enought to achieve best throughput.

  *********************************************** */   

#include "mublas_dot.def"

/* ***********************************************
    Macro
   *********************************************** */   
#if USE_NC_X_DOT
#define RESTRICT_X __restrict__
#define LOAD1_X LOAD1_NC
#define LOAD2_X LOAD2_NC
#define LOAD4_X LOAD4_NC
#else
#define RESTRICT_X
#define LOAD1_X LOAD1___
#define LOAD2_X LOAD2___
#define LOAD4_X LOAD4___
#endif

#if USE_NC_Y_DOT
#define RESTRICT_Y __restrict__
#define LOAD1_Y LOAD1_NC
#define LOAD2_Y LOAD2_NC
#define LOAD4_Y LOAD4_NC
#else
#define RESTRICT_Y
#define LOAD1_Y LOAD1___
#define LOAD2_Y LOAD2___
#define LOAD4_Y LOAD4___
#endif

#define NTX_DOT 128

/* ***********************************************
	Functions for determining Tx 
   *********************************************** */   
__host__ __inline__ void getBlkPerGrd_DOT ( mublasHandle_t* mh) {
	mh->Bx = ceilingEx ((float)mh->n / (float)(mh->N_T * mh->Tx), 1);
	mh->By = 1;
}

__host__ __inline__ void getShmPerBlk_DOT ( mublasHandle_t* mh) {
	mh->ShmPerBlk = ceilingEx ((float)(mh->Tx * mh->N_T * mh->N_one), mh->ShmUnit);
}

__host__ __inline__ size_t getTxSuf_DOT ( mublasHandle_t* mh) {
	return (size_t)ceilingEx ((float)mh->n / (float)mh->N_T / (float)mh->devprop.multiProcessorCount, mh->TxStep);
}


__global__ void
MUBLAS_SUM_DEVICE_s1 (
	SIZE_T n,
	TYPE0 *w
) {
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx + iTx;
	SIZE_T i;
	extern __shared__ TYPE0 Ts1[];
	TYPE1 Wr1, Tr1;

	Tr1 = ZERO1;
	if (rowid < n) {
		LOAD1_X (Wr1, &w[rowid]);
		Tr1 = ADD1 (Wr1, Tr1);
	}
	Ts1[iTx] = Tr1;
	__syncthreads ();

	#pragma unroll
	for (i = Tx/2; i >= WrpSize; i >>= 1) {
		if (iTx < i)
			Ts1[iTx] = ADD1 (Ts1[iTx], Ts1[iTx+i]);
		__syncthreads ();
	}

	Tr1 = Ts1[iTx];

	#pragma unroll
	for (i = WrpSize/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
	if (iTx == 0)
		w[iBx] = Tr1;
}

/* ***********************************************
	kernel=s1
   *********************************************** */   
__global__ void MUBLAS_DOT_DEVICE_s1 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 * RESTRICT_Y y,
	TYPE0 *w
) {
#if VEC1
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx + iTx;
	SIZE_T i;
	extern __shared__ TYPE1 Ts1[];
	TYPE1 Xr1, Yr1, Tr1;

	// Multiply-add
	Tr1 = ZERO1;
	if (rowid < n) {
		LOAD1_X (Xr1, &x[rowid]);
		LOAD1_Y (Yr1, &y[rowid]);
		Tr1 = MAD1 (Xr1, Yr1, Tr1);
	}
	Ts1[iTx] = Tr1;
	__syncthreads ();

	// Reduction
	#pragma unroll
	for (i = Tx/2; i >= WrpSize; i >>= 1) {
		if (iTx < i)
			Ts1[iTx] = ADD1 (Ts1[iTx], Ts1[iTx+i]);
		__syncthreads ();
	}

	Tr1 = Ts1[iTx];

	#pragma unroll
	for (i = WrpSize/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
	if (iTx == 0)
		w[iBx] = Tr1;
#endif
}

/* ***********************************************
	kernel=v2
   *********************************************** */   
__global__ void MUBLAS_DOT_DEVICE_v2 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 * RESTRICT_Y y,
	TYPE0 *w
) {
#if VEC2
	const SIZE_T N_T = 2;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx * N_T + iTx * N_T;
	SIZE_T i;
	extern __shared__ TYPE2 Ts2[];
	TYPE2 Xr2, Yr2, Tr2;
	TYPE1 Tr1;

	// Multiply-add
	Tr2 = ZERO2;
	if (rowid < n) {
		LOAD2_X (Xr2, &x[rowid]);
		LOAD2_Y (Yr2, &y[rowid]);
		Tr2 = MAD2 (Xr2, Yr2, Tr2);
	}
	Ts2[iTx] = Tr2;
	__syncthreads ();

	// Reduction
	#pragma unroll
	for (i = Tx/2; i >= WrpSize; i >>= 1) {
		if (iTx < i)
			Ts2[iTx] = ADD2 (Ts2[iTx], Ts2[iTx+i]);
		__syncthreads ();
	}

	Tr2 = Ts2[iTx];
	__syncthreads ();
#if defined PREC_C || defined PREC_B || defined PREC_DB
	Tr1 = ADD1 ((TYPE1&)Tr2.x, (TYPE1&)Tr2.z);
#else
	Tr1 = ADD1 (Tr2.x, Tr2.y);
#endif

	#pragma unroll
	for (i = WrpSize/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
	if (iTx == 0)
		w[iBx] = Tr1;
#endif
}
	
/* ***********************************************
	kernel=v4
   *********************************************** */   
__global__ void MUBLAS_DOT_DEVICE_v4 (
	const SIZE_T n,
	const TYPE0 * RESTRICT_X x,
	const TYPE0 * RESTRICT_Y y,
	TYPE0 *w
) {
#if VEC4
	const SIZE_T N_T = 4;
	const SIZE_T Tx = blockDim.x;
	const SIZE_T iTx = threadIdx.x;
	const SIZE_T iBx = blockIdx.x;
	const SIZE_T rowid = blockIdx.x * Tx * N_T + iTx * N_T;
	SIZE_T i;
	extern __shared__ TYPE4 Ts4[];
	TYPE4 Xr4, Yr4, Tr4;
	TYPE1 Tr1;

	// Multiply-add
	Tr4 = ZERO4;
	if (rowid < n) {
		LOAD4_X (Xr4, &x[rowid]);
		LOAD4_Y (Yr4, &y[rowid]);
		Tr4 = MAD4 (Xr4, Yr4, Tr4);
	}
	Ts4[iTx] = Tr4;
	__syncthreads ();

	// Reduction
	#pragma unroll
	for (i = Tx/2; i >= WrpSize; i >>= 1) {
		if (iTx < i)
			Ts4[iTx] = ADD4 (Ts4[iTx], Ts4[iTx+i]);
		__syncthreads ();
	}

	Tr4 = Ts4[iTx];
	__syncthreads ();
	Tr1 = ADD1(ADD1(Tr4.x, Tr4.y), ADD1(Tr4.z, Tr4.w));

	#pragma unroll
	for (i = WrpSize/2; i > 0; i >>= 1) 
		shuffleAdd (Tr1, i);
	if (iTx == 0)
		w[iBx] = Tr1;
#endif
}

#undef RESTRICT_X
#undef LOAD1_X
#undef LOAD2_X
#undef LOAD4_X
#undef RESTRICT_Y
#undef LOAD1_Y
#undef LOAD2_Y
#undef LOAD4_Y
