#ifndef PREC_B_H
#define PREC_B_H

#include "cudf.h"

#ifdef PREC_B
	// host
	#define MUBLAS_GEMV_MGPU_ASYNC mublasBgemvMgpuAsync
	#define MUBLAS_GEMV_MGPU mublasBgemvMgpu
	#define MUBLAS_CSRMV mublasBcsrmv
	#define MUBLAS_GEAM mublasBgeam
	#define MUBLAS_GEMM mublasBgemm
	#define MUBLAS_GEMV mublasBgemv
	#define MUBLAS_SYMV mublasBsymv
	#define MUBLAS_TRMV mublasBtrmv
	#define MUBLAS_AXPY mublasBaxpy
	#define MUBLAS_DOT  mublasBdot
	#define MUBLAS_SCAL mublasBscal
	#define MUBLAS_COPY mublasBcopy
//	#define CUBLAS_GEMM	
//	#define CUBLAS_GEMV 
//	#define CUBLAS_SYMV
//	#define CUBLAS_TRMV 
//	#define CUBLAS_AXPY 
//	#define CUBLAS_DOT
//	#define CUBLAS_SCAL 
//	#define CUBLAS_COPY 
	#define TYPE0 cudfreal
	#define ZERO0 make_cudfreal(0.,0.)
	#define ONE0 make_cudfreal(1.,0.)

	// vec1
	#define VEC1 1
	#define ZERO1 ZERO0
	#define TYPE1 TYPE0
	#define TYPEIN1 TYPE1
	#define ZEROIN1 ZERO1
	#define LOAD1___ ld_global_v2_f32
	#define LOAD1_NC ld_global_nc_v2_f32
	#define STORE1___ st_global_v2_f32

	#define ADD1 CUBADD
	#define MUL1 CUBMUL
	#define MAD1 CUBMAD

	// vec2
	#define VEC2 1
	#define TYPE2 cudfreal2
	#define ZERO2 make_cudfreal2(ZERO1,ZERO1)
	#define TYPEIN2 TYPE2
	#define ZEROIN2 ZERO2
	#define LOAD2___ ld_global_v4_f32
	#define LOAD2_NC ld_global_nc_v4_f32
	#define STORE2___ st_global_v4_f32

	#define ADD2 cuBadd2
	#define MUL12 cuBmul12
	#define MAD122 cuBmad122
	#define MAD212 cuBmad212
	#define MAD2 cuBmad2

	// vec4
	#define VEC4 0

	// DF-algorithm 
	#define CUBADD cuBaddV2
	#define CUBSUB cuBsubV2
	#define CUBMUL cuBmulV3
	#define CUBMAD cuBmuladd
#endif

typedef float4 cudfreal2;

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cudfreal &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = CUBADD (val, make_cudfreal (__shfl_xor(val.x, n, WrpSize), __shfl_xor(val.y, n, WrpSize)));
#endif
}
// =================================================================

// === Load ========================================================
__device__ __forceinline__ void ld_global_nc_v4_f32 (cudfreal2 &reg, const cudfreal *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v4.f32 {%0,%1,%2,%3}, [%4+0];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#else
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4+0];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#endif
#else
	reg = (cudfreal2&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_v2_f32 (cudfreal &reg, const cudfreal *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f32 {%0,%1}, [%2+0];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#else
	asm ("ld.global.v2.f32 {%0,%1}, [%2+0];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#endif
#else
	reg = *addr;
#endif
}

__device__ __forceinline__ void ld_global_v4_f32 (cudfreal2 &reg, const cudfreal *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4+0];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
#else
	reg = (cudfreal2&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_v2_f32 (cudfreal &reg, const cudfreal *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v2.f32 {%0,%1}, [%2+0];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
#else
	reg = *addr;
#endif
}
// =================================================================

// === Store =======================================================
__device__ __forceinline__ void st_global_v4_f32 (cudfreal *addr, const cudfreal2 reg) {
#if USE_INLINE_ASM
	asm ("st.global.v4.f32 [%0], {%1,%2,%3,%4};":: "l"(addr), "f"(reg.x), "f"(reg.y), "f"(reg.z), "f"(reg.w));
#else
	(cudfreal2&)*addr = reg;
#endif
}

__device__ __forceinline__ void st_global_v2_f32 (cudfreal *addr, const cudfreal reg) {
#if USE_INLINE_ASM
	asm ("st.global.v2.f32 [%0], {%1,%2};":: "l"(addr), "f"(reg.x), "f"(reg.y));
#else
	(cudfreal&)*addr = reg;
#endif
}
// =================================================================

__device__ __forceinline__ cudfreal2
make_cudfreal2
	(const cudfreal x, const cudfreal y)
{
	cudfreal2 v;
	v.x = x.x;
	v.y = x.y;
	v.z = y.x;
	v.w = y.y;
	return v;
}

__host__ __device__ __forceinline__ bool operator== (const cudfreal a, const cudfreal b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cudfreal
cuBmuladd
	(const cudfreal a, const cudfreal b, const cudfreal c)
{
	return CUBADD(CUBMUL(a, b), c);
}

__device__ __forceinline__ cudfreal2
cuBadd2
	(const cudfreal2 a, const cudfreal2 b)
{
	return make_cudfreal2 (CUBADD(make_cudfreal(a.x,a.y),make_cudfreal(b.x,b.y)), CUBADD(make_cudfreal(a.z,a.w),make_cudfreal(b.z,b.w)));
}

__device__ __forceinline__ cudfreal2
cuBmul12
	(const cudfreal a, const cudfreal2 b)
{
	return make_cudfreal2 (CUBMUL(a,make_cudfreal(b.x,b.y)), CUBMUL(a,make_cudfreal(b.z,b.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad2
	(const cudfreal2 a, const cudfreal2 b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(make_cudfreal(a.x,a.y),make_cudfreal(b.x,b.y),make_cudfreal(c.x,c.y)), CUBMAD(make_cudfreal(a.z,a.w),make_cudfreal(b.z,b.w),make_cudfreal(c.z,c.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad122
	(const cudfreal a, const cudfreal2 b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(a,make_cudfreal(b.x,b.y),make_cudfreal(c.x,c.y)), CUBMAD(a,make_cudfreal(b.z,b.w),make_cudfreal(c.z,c.w)));
}

__device__ __forceinline__ cudfreal2
cuBmad212
	(const cudfreal2 a, const cudfreal b, const cudfreal2 c)
{
	return make_cudfreal2 (CUBMAD(make_cudfreal(a.x,a.y),b,make_cudfreal(c.x,c.y)), CUBMAD(make_cudfreal(a.z,a.w),b,make_cudfreal(c.z,c.w)));
}

// === Texture =====================================================
texture<float2, cudaTextureType1D, cudaReadModeElementType> tex_mat_a;
texture<float2, cudaTextureType1D, cudaReadModeElementType> tex_mat_b;

static __forceinline__ __device__ cudfreal
tex_a(const int& i)
{
	float2 v = tex1Dfetch (tex_mat_a, i);
	return make_cudfreal (v.x, v.y);
}

static __forceinline__ __device__ cudfreal
tex_b(const int& i)
{
	float2 v = tex1Dfetch (tex_mat_b, i);
	return make_cudfreal (v.x, v.y);
}
// =================================================================

#endif
