#ifndef PREC_C_H
#define PREC_C_H

#ifdef PREC_C
	// host
	#define MUBLAS_GEMV_MGPU_ASYNC mublasCgemvMgpuAsync
	#define MUBLAS_GEMV_MGPU mublasCgemvMgpu
	#define MUBLAS_CSRMV mublasCcsrmv
	#define MUBLAS_GEAM mublasCgeam
	#define MUBLAS_GEMM mublasCgemm
	#define MUBLAS_GEMV mublasCgemv
	#define MUBLAS_HEMV mublasChemv
	#define MUBLAS_TRMV mublasCtrmv
	#define MUBLAS_AXPY mublasCaxpy
	#define MUBLAS_DOT  mublasCdot
	#define MUBLAS_SCAL mublasCscal
	#define MUBLAS_COPY mublasCcopy
	#define CUSPARSE_CSRMV cusparseCcsrmv
	#define CUBLAS_GEAM cublasCgeam
	#define CUBLAS_GEMM cublasCgemm
	#define CUBLAS_GEMV cublasCgemv
	#define CUBLAS_HEMV cublasChemv
	#define CUBLAS_TRMV cublasCtrmv
	#define CUBLAS_AXPY cublasCaxpy
	#define CUBLAS_DOT  cublasCdotu
	#define CUBLAS_SCAL cublasCscal
	#define CUBLAS_COPY cublasCcopy
	#define TYPE0 cuFloatComplex
	#define ZERO0 make_cuFloatComplex(0.,0.)
	#define ONE0 make_cuFloatComplex(1.,0.)

	// vec1
	#define VEC1 1
	#define ZERO1 ZERO0
	#define TYPE1 TYPE0
	#define ADD1 cuCaddf
	#define MUL1 cuCmulf
	#define MAD1 cuCmadf

	#define TYPEIN1 TYPE1
	#define ZEROIN1 ZERO1
	#define LOAD1___ ld_global_v2_f32
	#define LOAD1_NC ld_global_nc_v2_f32
	#define STORE1___ st_global_v2_f32

	// vec2
	#define VEC2 1
	#define TYPE2 cuFloatComplex2
	#define ZERO2 make_cuFloatComplex2(ZERO1,ZERO1)
	#define ADD2 cuCadd2
	#define MUL12 cuCmul12
	#define MAD122 cuCmad122
	#define MAD212 cuCmad212
	#define MAD2 cuCmad2

	#define TYPEIN2 TYPE2
	#define ZEROIN2 ZERO2
	#define LOAD2___ ld_global_v4_f32
	#define LOAD2_NC ld_global_nc_v4_f32
	#define STORE2___ st_global_v4_f32

	// vec4
	#define VEC4 0
#endif

typedef float4 cuFloatComplex2;

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cuFloatComplex &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = cuCaddf (val, make_cuFloatComplex (__shfl_xor(val.x, n, WrpSize), __shfl_xor(val.y, n, WrpSize)));
#endif
}
// =================================================================

// === Load ========================================================
__device__ __forceinline__ void ld_global_v4_f32 (cuFloatComplex2 &reg, const cuFloatComplex *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
#else
	reg = (cuFloatComplex2&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_v2_f32 (cuFloatComplex &reg, const cuFloatComplex *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v2.f32 {%0,%1}, [%2];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
#else
	reg = *addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_v4_f32 (cuFloatComplex2 &reg, const cuFloatComplex *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#else
	asm ("ld.global.v4.f32 {%0,%1,%2,%3}, [%4];": "=f"(reg.x), "=f"(reg.y), "=f"(reg.z), "=f"(reg.w): "l"(addr));
	#endif
#else
	reg = (cuFloatComplex2&)*addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_v2_f32 (cuFloatComplex &reg, const cuFloatComplex *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f32 {%0,%1}, [%2];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#else
	asm ("ld.global.v2.f32 {%0,%1}, [%2];": "=f"(reg.x), "=f"(reg.y): "l"(addr));
	#endif
#else
	reg = *addr;
#endif
}
// =================================================================

// === Store =======================================================
__device__ __forceinline__ void st_global_v4_f32 (cuFloatComplex *addr, const cuFloatComplex2 reg) {
#if USE_INLINE_ASM
	asm ("st.global.v4.f32 [%0], {%1,%2,%3,%4};":: "l"(addr), "f"(reg.x), "f"(reg.y), "f"(reg.z), "f"(reg.w));
#else
	(cuFloatComplex2&)*addr = reg;
#endif
}

__device__ __forceinline__ void st_global_v2_f32 (cuFloatComplex *addr, const cuFloatComplex reg) {
#if USE_INLINE_ASM
	asm ("st.global.v2.f32 [%0], {%1,%2};":: "l"(addr), "f"(reg.x), "f"(reg.y));
#else
	(cuFloatComplex&)*addr = reg;
#endif
}
// =================================================================

// === Vector ======================================================
__device__ __forceinline__ cuFloatComplex2
make_cuFloatComplex2
	(const cuFloatComplex x, const cuFloatComplex y)
{
	cuFloatComplex2 v;
	v.x = x.x;
	v.y = x.y;
	v.z = y.x;
	v.w = y.y;
	return v;
}

__host__ __device__ __forceinline__ bool operator== (const cuFloatComplex a, const cuFloatComplex b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cuFloatComplex
cuCmadf
	(const cuFloatComplex a, const cuFloatComplex b, const cuFloatComplex c)
{
	return cuCaddf(cuCmulf(a, b), c);
}

__device__ __forceinline__ cuFloatComplex2
cuCadd2
	(const cuFloatComplex2 a, const cuFloatComplex2 b)
{
	cuFloatComplex2 c;
	(cuFloatComplex&)c.x = cuCaddf (make_cuFloatComplex (a.x, a.y), make_cuFloatComplex (b.x, b.y));
	(cuFloatComplex&)c.z = cuCaddf (make_cuFloatComplex (a.z, a.w), make_cuFloatComplex (b.z, b.w));
	return c;
}

__device__ __forceinline__ cuFloatComplex2
cuCmul12
	(const cuFloatComplex a, const cuFloatComplex2 b)
{
	cuFloatComplex2 c;
	(cuFloatComplex&)c.x = cuCmulf (a, make_cuFloatComplex (b.x, b.y));
	(cuFloatComplex&)c.z = cuCmulf (a, make_cuFloatComplex (b.z, b.w));
	return c;
}

__device__ __forceinline__ cuFloatComplex2
cuCmad2
	(const cuFloatComplex2 a, const cuFloatComplex2 b, const cuFloatComplex2 c)
{
	cuFloatComplex2 d;
	(cuFloatComplex&)d.x = cuCaddf (cuCmulf (make_cuFloatComplex (a.x, a.y), make_cuFloatComplex (b.x, b.y)), make_cuFloatComplex (c.x, c.y));
	(cuFloatComplex&)d.z = cuCaddf (cuCmulf (make_cuFloatComplex (a.z, a.w), make_cuFloatComplex (b.z, b.w)), make_cuFloatComplex (c.z, c.w));
	return d;
}

__device__ __forceinline__ cuFloatComplex2
cuCmad122
	(const cuFloatComplex a, const cuFloatComplex2 b, const cuFloatComplex2 c)
{
	cuFloatComplex2 d;
	(cuFloatComplex&)d.x = cuCaddf (cuCmulf (a, make_cuFloatComplex (b.x, b.y)), make_cuFloatComplex (c.x, c.y));
	(cuFloatComplex&)d.z = cuCaddf (cuCmulf (a, make_cuFloatComplex (b.z, b.w)), make_cuFloatComplex (c.z, c.w));
	return d;
}

__device__ __forceinline__ cuFloatComplex2
cuCmad212
	(const cuFloatComplex2 a, const cuFloatComplex b, const cuFloatComplex2 c)
{
	cuFloatComplex2 d;
	(cuFloatComplex&)d.x = cuCaddf (cuCmulf (make_cuFloatComplex (a.x, a.y), b), make_cuFloatComplex (c.x, c.y));
	(cuFloatComplex&)d.z = cuCaddf (cuCmulf (make_cuFloatComplex (a.z, a.w), b), make_cuFloatComplex (c.z, c.w));
	return d;
}
// =================================================================

#endif
