#ifndef PREC_Z_H
#define PREC_Z_H

#ifdef PREC_Z
	// host
	#define MUBLAS_GEMV_MGPU_ASYNC mublasZgemvMgpuAsync
	#define MUBLAS_GEMV_MGPU mublasZgemvMgpu
	#define MUBLAS_CSRMV mublasZcsrmv
	#define MUBLAS_GEAM mublasZgeam
	#define MUBLAS_GEMM mublasZgemm
	#define MUBLAS_GEMV mublasZgemv
	#define MUBLAS_HEMV mublasZhemv
	#define MUBLAS_TRMV mublasZtrmv
	#define MUBLAS_AXPY mublasZaxpy
	#define MUBLAS_DOT  mublasZdot
	#define MUBLAS_SCAL mublasZscal
	#define MUBLAS_COPY mublasZcopy
	#define CUSPARSE_CSRMV cusparseZcsrmv
	#define CUBLAS_GEAM cublasZgeam
	#define CUBLAS_GEMM cublasZgemm
	#define CUBLAS_GEMV cublasZgemv
	#define CUBLAS_HEMV cublasZhemv
	#define CUBLAS_TRMV cublasZtrmv
	#define CUBLAS_AXPY cublasZaxpy
	#define CUBLAS_DOT  cublasZdotu
	#define CUBLAS_SCAL cublasZscal
	#define CUBLAS_COPY cublasZcopy
	#define TYPE0 cuDoubleComplex
	#define ZERO0 make_cuDoubleComplex(0.,0.)
	#define ONE0 make_cuDoubleComplex(1.,0.)

	// vec1
	#define VEC1 1
	#define TYPE1 TYPE0
	#define ZERO1 ZERO0
	#define ADD1 cuCadd
	#define MUL1 cuCmul
	#define MAD1 cuCmad

	#define TYPEIN1 TYPE1
	#define ZEROIN1 ZERO1
	#define LOAD1___ ld_global_v2_f64
	#define LOAD1_NC ld_global_nc_v2_f64
	#define STORE1___ st_global_v2_f64

	// vec2
	#define VEC2 0

	// vec4
	#define VEC4 0
#endif

// === Shuffle =====================================================
__device__ __forceinline__ void shuffleAdd (cuDoubleComplex &val, const int n) {
#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 300)
	val = cuCadd (val, make_cuDoubleComplex ( 
		__hiloint2double( __shfl_xor(__double2hiint(val.x), n, WrpSize), __shfl_xor(__double2loint(val.x), n, WrpSize)),
		__hiloint2double( __shfl_xor(__double2hiint(val.y), n, WrpSize), __shfl_xor(__double2loint(val.y), n, WrpSize))));
#endif
}
// =================================================================

// === Load ========================================================
__device__ __forceinline__ void ld_global_v2_f64 (cuDoubleComplex &reg, const cuDoubleComplex *addr) {
#if USE_INLINE_ASM
	asm ("ld.global.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x), "=d"(reg.y): "l"(addr));
#else
	reg = *addr;
#endif
}

__device__ __forceinline__ void ld_global_nc_v2_f64 (cuDoubleComplex &reg, const cuDoubleComplex *addr) {
#if USE_INLINE_ASM
	#if defined (__CUDA_ARCH__) && (__CUDA_ARCH__ >= 320)
	asm ("ld.global.nc.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x), "=d"(reg.y): "l"(addr));
	#else
	asm ("ld.global.v2.f64 {%0,%1}, [%2+0];": "=d"(reg.x), "=d"(reg.y): "l"(addr));
	#endif
#else
	reg = *addr;
#endif
}
// =================================================================

// === Store =======================================================
__device__ __forceinline__ void st_global_v2_f64 (cuDoubleComplex *addr, const cuDoubleComplex reg) {
#if USE_INLINE_ASM
	asm ("st.global.v2.f64 [%0], {%1,%2};":: "l"(addr), "d"(reg.x), "d"(reg.y));
#else
	*addr = reg;
#endif
}
// =================================================================

__host__ __device__ __forceinline__ bool operator== (const cuDoubleComplex a, const cuDoubleComplex b) {
	if ((a.x == b.x) && (a.y == b.y))
		return true;
	else
		return false;
}

__device__ __forceinline__ cuDoubleComplex
cuCmad
	(const cuDoubleComplex a, const cuDoubleComplex b, const cuDoubleComplex c)
{
	return cuCadd(cuCmul(a, b), c);
}

#endif
