/*
	// temperature evaluation code using NVML
	// (deprecated in mublas-1.5.2)
	// (1) initialization
	#ifdef USE_NVML
	uint32_t temp1 = 0, temp2 = 0;
	nvmlDevice_t dev;
	nvmlInit();
	nvmlDeviceGetHandleByIndex(th.devno, &dev);
	do {
		sleep(1);
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
	} while (temp1 >= TEMPLIMIT);
	#endif

	// (2) temperature evaluation
	#if defined (USE_NVML)
	nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp2);
	#endif

	// (3) shutdown
	#if defined (USE_NVML)
	nvmlShutdown();
	#endif
*/

#define NGPU_MAX 8

// ----- testing handler ---------------------------------------------------------------
struct testingHandle_t {
	// multi-GPU
	uint32_t num_gpus;
	uint32_t num_installed_gpus;
	uint32_t devno;

	// device parameter (single GPU only)
	cudaDeviceProp devprop[NGPU_MAX];
	int32_t cc[NGPU_MAX];
	double device_bytes_per_sec[NGPU_MAX];
	double device_flops_per_sec_sp[NGPU_MAX];
	double device_flops_per_sec_dp[NGPU_MAX];
	int32_t Cores[NGPU_MAX];
	int32_t DPCores[NGPU_MAX];

	// testing parameter
	char mode; // c=check, p=perf

	double routine_flops_per_sec_theoretical_peak;
	double routine_flops;
	double routine_bytes;

	size_t TxStatic;
	size_t TyStatic;
	int32_t driverVersion; 
	int32_t runtimeVersion;
	size_t test_points;
	int32_t kernelnumber;
	int32_t kerneldebug;
	int32_t cublasatomic;
	int32_t nodisp;

	// BLAS parameters
	size_t word_size;
	// dim
	size_t kl;
	size_t ku;
	size_t dim_start;
	size_t dim_stop;
	size_t dim_step;
	size_t dim_m_const;
	size_t dim_n_const;
	size_t dim_k_const;
	size_t dim_m_dev;
	size_t dim_n_dev;
	size_t dim_k_dev;
	size_t dim_m_hst;
	size_t dim_n_hst;
	size_t dim_k_hst;
	size_t incx;
	size_t incy;
	uint32_t mode_dim_start;
	uint32_t mode_dim_stop;
	uint32_t dim_step_mode;

	// options
	char tranA; // N, T, C
	char tranB; // N, T, C
	char uplo; // U, L
	char diag; // N, U
	char side; // L, R

	// sparse matrix
	char mtx_file[256];
};

// ----- CUBLAS const value function ---------------------------------------------------------------
inline cublasOperation_t cublas_trans_const_m (char tran) {
	if (tran == 'N' || tran == 'n') return CUBLAS_OP_N;
	if (tran == 'T' || tran == 't') return CUBLAS_OP_T;
	if (tran == 'C' || tran == 'c') return CUBLAS_OP_C;
	return CUBLAS_OP_N; //default
}
inline cublasFillMode_t cublas_uplo_const_m (char uplo) {
	if (uplo == 'L' || uplo == 'l') return CUBLAS_FILL_MODE_LOWER;
	if (uplo == 'U' || uplo == 'u') return CUBLAS_FILL_MODE_UPPER;
	return CUBLAS_FILL_MODE_LOWER; //default
}
inline cublasDiagType_t cublas_diag_const_m (char diag) {
	if (diag == 'N' || diag == 'n') return CUBLAS_DIAG_NON_UNIT;
	if (diag == 'U' || diag == 'u') return CUBLAS_DIAG_UNIT;
	return CUBLAS_DIAG_NON_UNIT; //default
}
inline cublasSideMode_t cublas_side_const_m (char side) {
	if (side == 'L' || side == 'l') return CUBLAS_SIDE_LEFT;
	if (side == 'R' || side == 'r') return CUBLAS_SIDE_RIGHT;
	return CUBLAS_SIDE_LEFT; //default
}

// ----- CUSPARSE const value function ---------------------------------------------------------------
#if defined (CSRMV) || defined (HYBMV)
inline cusparseOperation_t cusparse_trans_const_m (char tran) {
	if (tran == 'N' || tran == 'n') return CUSPARSE_OPERATION_NON_TRANSPOSE;
	if (tran == 'T' || tran == 't') return CUSPARSE_OPERATION_TRANSPOSE;
	if (tran == 'C' || tran == 'c') return CUSPARSE_OPERATION_CONJUGATE_TRANSPOSE;
	return CUSPARSE_OPERATION_NON_TRANSPOSE; //default
}
#endif

// ----- CBLAS const value function ---------------------------------------------------------------
#if defined (CHECK_CBLAS)
inline CBLAS_TRANSPOSE cblas_trans_const_m (char tran) {
	if (tran == 'N' || tran == 'n') return CblasNoTrans;
	if (tran == 'T' || tran == 't') return CblasTrans;
	if (tran == 'C' || tran == 'c') return CblasConjTrans;
	return CblasNoTrans; //default
}
inline CBLAS_UPLO cblas_uplo_const_m (char uplo) {
	if (uplo == 'L' || uplo == 'l') return CblasLower;
	if (uplo == 'U' || uplo == 'u') return CblasUpper;
	return CblasLower; //default
}
inline CBLAS_DIAG cblas_diag_const_m (char diag) {
	if (diag == 'N' || diag == 'n') return CblasNonUnit;
	if (diag == 'U' || diag == 'u') return CblasUnit;
	return CblasNonUnit; //default
}
inline CBLAS_SIDE cblas_side_const_m (char side) {
	if (side == 'L' || side == 'l') return CblasLeft;
	if (side == 'R' || side == 'r') return CblasRight;
	return CblasLeft; //default
}
#endif

// ----- MBLAS compatible ---------------------------------------------------------------
#if defined (MUBLAS) && defined (CHECK_MPACK)
inline cuddreal make_cuddreal (const qd_real v) {
	return make_cuddreal (v.x[0], v.x[1]);
}

inline cuddcomplex make_cuddcomplex (const qd_real r, const qd_real i) {
	return make_cuddcomplex (make_cuddreal (r), make_cuddreal (i));
}
#endif

#if defined (MUBLAS)
void set_mublas_handler (testingHandle_t *th, mublasHandle_t *mh) {
	mh->TxStatic = th->TxStatic;
	mh->TyStatic = th->TyStatic;
	mh->kernelnumber = th->kernelnumber;
	mh->kerneldebug = th->kerneldebug;
	if (th->mode == 't') mh->autotuning = 1;
}
#endif

void get_routine_theoretial_performance (testingHandle_t *th) {
	size_t s;
	s = th->word_size;
	size_t n;
	n = th->dim_n_dev;
#if defined (DIM2) || defined (DIM3)
	size_t m;
	m = th->dim_m_dev;
#endif
#if defined (DIM3)
	size_t k;
	k = th->dim_k_dev;
#endif
#if defined (DIM1)
	th->routine_flops = (double) ROUTINE_FLOPS (n);
	th->routine_bytes = (double) ROUTINE_BYTES (n, s);
#elif defined (DIM2) 
	th->routine_flops = (double) ROUTINE_FLOPS (m, n);
	th->routine_bytes = (double) ROUTINE_BYTES (m, n, s);
#elif defined (DIM3) 
	th->routine_flops = (double) ROUTINE_FLOPS (m, n, k);
	th->routine_bytes = (double) ROUTINE_BYTES (m, n, k, s);
#endif
}

void dim_hst_setup (testingHandle_t *th) {
	th->dim_m_hst = th->dim_stop;
	th->dim_n_hst = th->dim_stop;
	th->dim_k_hst = th->dim_stop;
	if (th->dim_m_const > 0) th->dim_m_hst = th->dim_m_const;
	if (th->dim_n_const > 0) th->dim_n_hst = th->dim_n_const;
	if (th->dim_k_const > 0) th->dim_k_hst = th->dim_k_const;
#if defined (DIM1)
	if (th->dim_m_const > 0) {
		th->dim_start = th->dim_m_const;
		th->dim_stop  = th->dim_m_const;
		th->dim_n_hst = th->dim_m_const;
	}
	if (th->dim_n_const > 0) {
		th->dim_start = th->dim_n_const;
		th->dim_stop  = th->dim_n_const;
		th->dim_n_hst = th->dim_n_const;
	}
#endif
}

void dim_dev_setup (testingHandle_t *th) {
	th->dim_m_dev = th->dim_start;
	th->dim_n_dev = th->dim_start;
	th->dim_k_dev = th->dim_start;
	if (th->dim_m_const > 0) th->dim_m_dev = th->dim_m_const;
	if (th->dim_n_const > 0) th->dim_n_dev = th->dim_n_const;
	if (th->dim_k_const > 0) th->dim_k_dev = th->dim_k_const;
}

void dim_dev_increment (testingHandle_t *th) {
	// *= 2: log mode
	// += th->dim_step: normal mode
	if (th->dim_m_const == 0) {
		if (th->dim_step_mode == 2)
			th->dim_m_dev *= 2;
		else
			th->dim_m_dev += th->dim_step;
	}
	if (th->dim_n_const == 0) {
		if (th->dim_step_mode == 2)
			th->dim_n_dev *= 2;
		else
			th->dim_n_dev += th->dim_step;
	}
	if (th->dim_k_const == 0) {
		if (th->dim_step_mode == 2)
			th->dim_k_dev *= 2;
		else
			th->dim_k_dev += th->dim_step;
	}
}

double gettime () {
	struct timeval tv;
	cudaDeviceSynchronize ();
	gettimeofday (&tv, NULL);
	return tv.tv_sec + (double) tv.tv_usec * 1.0e-6;
}

int32_t get_num_sp_cores (int32_t cc) {
	switch (cc) {
		case 100: return 8;
		case 110: return 8;
		case 120: return 8;
		case 130: return 8;
		case 200: return 32;
		case 210: return 48;
		case 300: return 192;
		case 320: return 192;
		case 350: return 192;
		case 370: return 192;
		case 500: return 128;
		case 520: return 128;
		default: 
			printf ("error: unknown architecture\n");
			return 0;
	}
}

int32_t get_num_dp_cores (int32_t cc) {
	switch (cc) {
		case 100: return 1;
		case 110: return 1;
		case 120: return 1;
		case 130: return 1;
		case 200: return 16;
		case 210: return 0; //?
		case 300: return 64;
		case 320: return 64;
		case 350: return 64;
		case 370: return 64;
		case 500: return 0;
		case 520: return 4;
		default: 
			printf ("error: unknown architecture\n");
			return 0;
	}
}

void testingCreate (int32_t argc, char **argv, testingHandle_t *th) {
	int32_t const_m = 0, const_n = 0, const_k = 0;

 	// default ---------------------------
	th->num_gpus = 1;
	th->devno = 0;
	th->mode = 'c';
	// options
	th->tranA = 'N';
	th->tranB = 'N';
	th->uplo = 'L';
	th->diag = 'N';
	th->side = 'L';

	th->TxStatic = 0;
	th->TyStatic = 0;
	th->kernelnumber = 0;
	th->kerneldebug = 0;
	th->cublasatomic = 0;
	th->nodisp = 0;

	th->dim_start	= 128;
	th->dim_stop	= 1024;
	th->dim_step	= 128;
	th->dim_m_dev = 0;
	th->dim_n_dev = 0;
	th->dim_k_dev = 0;
	th->dim_m_hst = 0;
	th->dim_n_hst = 0;
	th->dim_k_hst = 0;
	th->dim_m_const = 0;
	th->dim_n_const = 0;
	th->dim_k_const = 0;
	th->incx = 1;
	th->incy = 1;
	th->test_points = 0;
	th->mode_dim_start = 0;
	th->mode_dim_stop = 0;
	th->dim_step_mode = 0;
	th->word_size = sizeof(GPU_TYPE);
	strcpy (th->mtx_file, "");

 	// args ---------------------------
	int32_t opt, opt_index;
	optind = 1;
	struct option long_options[] = {
		{"numgpus", required_argument, NULL, 'c'},
		{"devno", required_argument, NULL, 'g'},
		{"m", required_argument, NULL, 'm'},
		{"n", required_argument, NULL, 'n'},
		{"k", required_argument, NULL, 'k'},
		{"range", required_argument, NULL, 'r'},
		{"trans", required_argument, 0, 'a'},
		{"tranb", required_argument, 0, 'b'},
		{"uplo", required_argument, 0, 'u'},
		{"diag", required_argument, 0, 'd'},
		{"side", required_argument, 0, 's'},
		{"mode", required_argument, NULL, 'o'},
		{"kernel", required_argument, 0, 'v'},
		{"atomic", no_argument, NULL, 't'},
		{"tx", required_argument, NULL, 'x'},
		{"ty", required_argument, NULL, 'y'},
		{"points", required_argument, NULL, 'p'},
		{"help", no_argument, NULL, 'h'},
		{"incx", required_argument, NULL, 'i'},
		{"incy", required_argument, NULL, 'j'},
		{"nodisp", required_argument, NULL, 'w'},
		{"mtx", required_argument, NULL, 'f'},
		{0, 0, 0, 0}
	};
	while ((opt = getopt_long (argc, argv, "cgmnkrabudsovtxyrphijwf", long_options, &opt_index)) != -1){
		switch (opt) {
			// dim
			case 'c':{
				th->num_gpus = atoi(optarg);
				break;
			} case 'g':{
				th->devno = atoi(optarg);
				break;
			} case 'm':{
				th->dim_m_const = (size_t)atoi(optarg);
				const_m = 1;
				break;
			} case 'n':{
				th->dim_n_const = (size_t)atoi(optarg);
				const_n = 1;
				break;
			} case 'k':{
				th->dim_k_const = (size_t)atoi(optarg);
				const_k = 1;
				break;
			} case 'r':{
				char *str, *c_start, *c_stop, *c_step;
				char par[] = ":";
				str = optarg;
				c_start = strtok (str, par);
				c_stop  = strtok (NULL, par);
				c_step	= strtok (NULL, par);

				if (strcmp (c_start, "max") == 0)
					th->mode_dim_start	= 1;
				else 
					th->dim_start	= atoi (c_start);

				if (strcmp (c_stop, "max") == 0)
					th->mode_dim_stop	= 1;
				else 
					th->dim_stop	= atoi (c_stop);

				if (strcmp (c_step, "log") == 0)
					th->dim_step_mode	= 2;
				else 
					th->dim_step	= atoi (c_step);

				break;
			} case 'a':{
			// options
				// N, T, C
				if (optarg[0]) th->tranA= optarg[0];
				break;
			} case 'b':{
				// N, T, C
				if (optarg[0]) th->tranB= optarg[0];
				break;
			} case 'u':{
				// U, L
				if (optarg[0]) th->uplo = optarg[0];
				break;
			} case 'd':{
				// N, U
				if (optarg[0]) th->diag = optarg[0];
				break;
			} case 's':{
				// L, R
				if (optarg[0]) th->side = optarg[0];
				break;
			} case 'o':{
				th->mode = optarg[0];
				break;
			} case 'v':{
				if (optarg) th->kerneldebug = atoi(optarg);
				break;
			} case 't':{
				th->cublasatomic = 1;
				break;
			} case 'x':{
				th->TxStatic = atoi(optarg);
				break;
			} case 'y':{
				th->TyStatic = atoi(optarg);
				break;
			} case 'p':{
				th->test_points = atoi(optarg);
				break;
			} case 'h':{
				printf ("testing program:\n");
				break;
			} case 'i':{
				th->incx = atoi(optarg);
				break;
			} case 'j':{
				th->incy = atoi(optarg);
				break;
			} case 'w':{
				th->nodisp = atoi(optarg);
				break;
			} case 'f':{
				strcpy (th->mtx_file, optarg);
				char buf[256];
				FILE *fp;
				if ((fp = fopen (th->mtx_file, "r")) == NULL) {
					printf ("mublas testing error: file open error.\n");
					exit (1);
				}	
				while (fgets (buf, 256, fp) != NULL) {
					if (buf[0] == '%') {
						if ((strstr (buf, "symmetric")) != NULL) {
							printf ("mublas testing error: symmetric matrix is not supported on this version.");
							exit (1);
						}
					} else {
						sscanf (buf, "%zd%zd%zd", &th->dim_m_dev, &th->dim_n_dev, &th->dim_k_dev);
						if (th->dim_m_dev <= 0 || th->dim_n_dev <= 0 || th->dim_k_dev <= 0) {
							printf ("mublas testing error: illegal matrix size.");
							exit (1);
						}
						goto end_of_case_f;
						break;
					}
				}
				end_of_case_f:
					fclose (fp);
					break;
			}
		}
	}


// ----- init devinfo -----
	int32_t driverVersion = 0;
	int32_t runtimeVersion = 0;
	int32_t devcnt;
	cudaGetDeviceCount (&devcnt);
	th->num_installed_gpus = (size_t)devcnt;
	cudaDeviceProp devprop;  
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);
	th->driverVersion = driverVersion;
	th->runtimeVersion = runtimeVersion;
	for (uint32_t igpu = th->devno; igpu < std::max (th->num_gpus, th->devno+1); igpu++) {
		cudaSetDevice (igpu);
		cudaGetDeviceProperties (&devprop, igpu);
		th->devprop[igpu] = devprop;
		th->cc[igpu] = devprop.major * 100 + devprop.minor * 10;
		th->Cores[igpu] = get_num_sp_cores (th->cc[igpu]);
		th->DPCores[igpu] = get_num_dp_cores (th->cc[igpu]);
		th->device_bytes_per_sec[igpu] = (double)th->devprop[igpu].memoryClockRate * 2. * 1000. * (double)th->devprop[igpu].memoryBusWidth / 8.;
		th->device_flops_per_sec_sp[igpu] = (double)th->devprop[igpu].clockRate * 1000. * 2. * (double)th->devprop[igpu].multiProcessorCount * th->Cores[igpu];
		th->device_flops_per_sec_dp[igpu] = (double)th->devprop[igpu].clockRate * 1000. * 2. * (double)th->devprop[igpu].multiProcessorCount * th->DPCores[igpu];
	}
#if defined (PREC_S) || defined (PREC_C) // single precision
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_sp[th->devno];
#elif defined (PREC_D) || defined (PREC_Z) // double precision
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_dp[th->devno];
#elif defined (PREC_B) || defined (PREC_DF) || defined (PREC_V) // DF-operation
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_sp[th->devno] / (double)DFMADCOUNT;
#elif defined (PREC_W) || defined (PREC_U) // DD-operation
	th->routine_flops_per_sec_theoretical_peak = th->device_flops_per_sec_dp[th->devno] / (double)DDMADCOUNT;
#else // unknown
	th->routine_flops_per_sec_theoretical_peak = 0.;
#endif

// ----- problem setting -----
	size_t max_size = 0;
	#if defined (DIM2) || defined (DIM3)
	size_t dim_1;
	#endif
	#if defined (DIM3)
	size_t dim_2;
	#endif

	switch (const_m + const_n + const_k) {
		case 0: {
			max_size = ROUTINE_MAXSIZE (th->devprop[th->devno].totalGlobalMem, MEMMARGIN, th->word_size);
			break;
		}
		#if defined (DIM2) || defined (DIM3)
		case 1: {
			dim_1 = th->dim_m_const + th->dim_n_const + th->dim_k_const;
			max_size = ROUTINE_MAXSIZE_1 (th->devprop[th->devno].totalGlobalMem, MEMMARGIN, th->word_size, dim_1);
			break;
		}
		#endif
		#if defined (DIM3)
		case 2: {
			if (th->dim_m_const == 0) {
				dim_1 = th->dim_n_const;
				dim_2 = th->dim_k_const;
			} else if (th->dim_n_const == 0) {
				dim_1 = th->dim_m_const;
				dim_2 = th->dim_k_const;
			} else if (th->dim_k_const == 0) {
				dim_1 = th->dim_n_const;
				dim_2 = th->dim_m_const;
			}
			max_size = ROUTINE_MAXSIZE_2 (th->devprop[th->devno].totalGlobalMem, MEMMARGIN, th->word_size, dim_1, dim_2);
			break;
		}
		#endif
	}

	if (th->mode_dim_start == 1) th->dim_start = max_size;
	if (th->mode_dim_stop  == 1) th->dim_stop = max_size;
	if (th->test_points != 0) {
		size_t ccc = 128, ttt = 1;
		ttt = (th->dim_stop/(th->test_points)/ccc)*ccc;
		while (ttt < 1 && ccc > 0) {
			ccc /= 2;
			ttt = (th->dim_stop/(th->test_points)/ccc)*ccc;
		}
		if (ttt < 1) ttt = 1;
		th->dim_step = ttt;
		th->dim_start = th->dim_step;
	}
	
	th->test_points = (th->dim_stop-th->dim_start)/th->dim_step;
// ---------------------
}

int32_t get_num_iter (testingHandle_t *th) {
	return std::min ((double)MAXLOOP, std::max ((double)MINLOOP, MINSEC / std::max
	(th->routine_flops / th->routine_flops_per_sec_theoretical_peak, th->routine_bytes / th->device_bytes_per_sec[th->devno])));
}

// DF-operation
#if defined (MUBLAS)
#define SPLITTER 536870913.0 // 2^29+1
cudfreal double_to_cudfreal (const double d) {
	double t, h, l;
	t = SPLITTER * d;
	h = t - (t - d);
	l = d - h;
	return make_cudfreal ((float)h, (float)l);
}

double cudfreal_to_double (const cudfreal df) {
	return (double)df.x + (double)df.y;
}
#endif

#if (defined (CHECK_MPACK)) && (defined (PREC_W) || defined (PREC_U))
#define RAND qdrand()/((REFERENCE_BASETYPE)RAND_MAX+1.)
#else
#define RAND rand()/((REFERENCE_BASETYPE)RAND_MAX+1.)
#endif

void mublasInit2D (size_t rd, size_t cd, size_t ld, GPU_TYPE *mat) {
	size_t i, j;
	int32_t sig = 1;
	if (ld == 0) ld = rd;
	for (j = 0; j < cd; j++) {
		for (i = 0; i < rd; i++) {
			if (rand()%2 == 0) sig *= -1;
			REFERENCE_BASETYPE r1 = sig*RAND;
#if defined (PREC_C) || defined (PREC_Z) || defined (PREC_U) || defined (PREC_V)
			if (rand()%2 == 0) sig *= -1;
			REFERENCE_BASETYPE r2 = sig*RAND;
			#if defined (PREC_U) || !defined (CHECK_MPACK)
			mat[j*ld+i] = MAKE_GPU_TYPE (r1, r2);
			#else
			mat[j*ld+i] = MAKE_GPU_TYPE (r1.x[0], r2.x[0]);
			#endif
#else
			#if defined (PREC_W) || !defined (CHECK_MPACK)
			mat[j*ld+i] = MAKE_GPU_TYPE (r1);
			#else
			mat[j*ld+i] = MAKE_GPU_TYPE (r1.x[0]);
			#endif
#endif
		}
	}
	// Hermitian (for MAGMA)
	if (cd > 1) {
		for (i = 0; i < rd; i++) {
			#if defined (PREC_C)
			mat[i*ld+i] = make_cuFloatComplex(mat[i*ld+i].x,0);
			#elif defined (PREC_Z)
			mat[i*ld+i] = make_cuDoubleComplex(mat[i*ld+i].x,0);
			#endif
		}
	}
}

#if defined (PREC_S)
void mublasInit2Dzero (size_t rd, size_t cd, size_t ld, GPU_TYPE *mat) {
	size_t i, j;
	if (ld == 0) ld = rd;
	for (j = 0; j < cd; j++) {
		for (i = 0; i < rd; i++) {
			mat[j*ld+i] = 0.;
		}
	}
}

void mublasInit2Done (size_t rd, size_t cd, size_t ld, GPU_TYPE *mat) {
	size_t i, j;
	if (ld == 0) ld = rd;
	for (j = 0; j < cd; j++) {
		for (i = 0; i < rd; i++) {
			mat[j*ld+i] = 1.;
		}
	}
}
#endif

/*
// under construction...

void mublasInitCSRMatrix (testingHandle_t *th, GPU_TYPE *hst_Aval, int32_t *hst_Aptr, int32_t *hst_Aind) {
	size_t i, j;
	int32_t sym = 0;
	int32_t *Arow = NULL, *Acol = NULL;
	double *Aval = NULL;
	char buf[256];
	FILE *fp;

	if ((fp = fopen (th->mtx_file, "r")) == NULL) {
		printf ("mublas testing error: file open error.\n");
		exit (1);
	}
	// malloc matrix market format working area
	Arow = (int32_t *) malloc (sizeof(int32_t) * th->dim_k_dev);
	Acol = (int32_t *) malloc (sizeof(int32_t) * th->dim_k_dev);
	Aval = (double *) malloc (sizeof(double) * th->dim_k_dev);
	// read matrix market format
	i = 0;
	while (fgets (buf, 256, fp) != NULL) {
		if (buf[0] == '%') {
			if ((strstr (buf, "symmetric")) != NULL) sym = 1;
		} else {
			if (i != 0) {
				sscanf (buf, "%d%d%lf", &Arow[i-1], &Acol[i-1], &Aval[i-1]);
				printf ("Arow=%d, Acol=%d, Aval=%lf\n", Arow[i-1], Acol[i-1], Aval[i-1]);
			}
			i++;
		}
	}
	// sorting
	sort (Arow, Acol, Aval, th->dim_k_dev);
	// storing 
	for (i = 0; i < th->dim_n_dev; i++) {
		if (Arow[i] == i)
		else 
			hst_Aptr[i] = 0;
	}

	for (i = 0; i < th->dim_n_dev; i++) {
		hst_Arow[i] = 0;
	hst_Aval[0] = Aval[0];
	hst_Aind[0] = Acol[0];
	for (j = 0; j < Arow[0]; j++) 
		hst_Aptr[j] = 0;
	for (i = 1; i < th->dim_k_dev; i++) {
		hst_Aval[i] = Aval[i];
		hst_Aind[i] = Acol[i];
		if (i != 0 && Arow[i-1] < Arow[i]) {
			hst_Aptr[j] = i;
			j++;
		}
		if (i != 0 && Arow[i-1] <= Arow[i]) {
			for (k = 0; k < Arow[i]-Arow[i-1]; k++) {
				hst_Aptr[j] = 0;
				j++
			}
		}
	}
	hst_Arow[dim_n_dev] = th->dim_k_dev;

	fclose (fp);
	free (Arow);
	free (Acol);
	free (Aval);

	exit (0);
}
*/

#if defined (CHECK_MPACK)
void mublasConvert2D (size_t rd, size_t cd, size_t ld, GPU_TYPE *src, MBLAS_TYPE *dst) {
	size_t i, j;
	if (ld == 0) ld = rd;
	for (j = 0; j < cd; j++) {
		for (i = 0; i < rd; i++) {
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_DB)
			dst[j*ld+i] = qd_real ((double)src[j*ld+i], 0., 0., 0.);
#endif
#if defined (PREC_W)
			dst[j*ld+i] = qd_real (src[j*ld+i].x, src[j*ld+i].y, 0., 0.);
#endif
#if defined (PREC_C) || defined (PREC_Z) 
			dst[j*ld+i] = qd_complex (qd_real ((double)src[j*ld+i].x, 0., 0., 0.), qd_real ((double)src[j*ld+i].y, 0., 0., 0.));
#endif
#if defined (PREC_U) 
			dst[j*ld+i] = qd_complex (qd_real (src[j*ld+i].x.x, src[j*ld+i].x.y, 0., 0.), qd_real (src[j*ld+i].y.x, src[j*ld+i].y.y, 0., 0.));
#endif
#if defined (PREC_B)
			dst[j*ld+i] = qd_real (cudfreal_to_double (src[j*ld+i]), 0., 0., 0.);
#endif
#if defined (PREC_V) 
			dst[j*ld+i] = qd_complex (qd_real (cudfreal_to_double (src[j*ld+i].x), 0., 0., 0.), qd_real (cudfreal_to_double (src[j*ld+i].y), 0., 0., 0.));
#endif
		}
	}
}
#elif defined (CHECK_CBLAS)
void mublasConvert2D (size_t rd, size_t cd, size_t ld, GPU_TYPE *src, CBLAS_TYPE *dst) {
	size_t i, j;
	if (ld == 0) ld = rd;
	for (j = 0; j < cd; j++) {
		for (i = 0; i < rd; i++) {
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_DB)
			dst[j*ld+i] = src[j*ld+i];
#endif
#if defined (PREC_W)
			dst[j*ld+i] = src[j*ld+i].x;
#endif
#if defined (PREC_C) || defined (PREC_Z) 
			dst[j*ld+i].r = src[j*ld+i].x;
			dst[j*ld+i].i = src[j*ld+i].y;
#endif
#if defined (PREC_U) 
			dst[j*ld+i].r = src[j*ld+i].x.x;
			dst[j*ld+i].i = src[j*ld+i].y.x;
#endif
#if defined (PREC_B)
			dst[j*ld+i] = cudfreal_to_double (src[j*ld+i]);
#endif
#if defined (PREC_V)
			dst[j*ld+i].r = cudfreal_to_double (src[j*ld+i].x);
			dst[j*ld+i].i = cudfreal_to_double (src[j*ld+i].y);
#endif
		}
	}
}
#elif defined (CHECK_CUBLAS)
void mublasConvert2D (size_t rd, size_t cd, size_t ld, GPU_TYPE *src, CUBLAS_TYPE *dst) {
	size_t i, j;
	if (ld == 0) ld = rd;
	for (j = 0; j < cd; j++) {
		for (i = 0; i < rd; i++) {
#if defined (PREC_S) || defined (PREC_D) || defined (PREC_DB)
			dst[j*ld+i] = src[j*ld+i];
#endif
#if defined (PREC_W)
			dst[j*ld+i] = src[j*ld+i].x;
#endif
#if defined (PREC_C) || defined (PREC_Z) 
			dst[j*ld+i].x = src[j*ld+i].x;
			dst[j*ld+i].y = src[j*ld+i].y;
#endif
#if defined (PREC_U) 
			dst[j*ld+i].x = src[j*ld+i].x.x;
			dst[j*ld+i].y = src[j*ld+i].y.x;
#endif
#if defined (PREC_B)
			dst[j*ld+i] = cudfreal_to_double (src[j*ld+i]);
#endif
#if defined (PREC_V)
			dst[j*ld+i].x = cudfreal_to_double (src[j*ld+i].x);
			dst[j*ld+i].y = cudfreal_to_double (src[j*ld+i].y);
#endif
		}
	}
}
#endif

// check routine (|trg-ref|/|ref|) -------------------------------------------------
void mublasCheckVector (size_t m, GPU_TYPE *target, REFERENCE_TYPE *ref) {
	size_t size = sizeof (REFERENCE_TYPE);
	REFERENCE_TYPE negone = REFERENCE_NEGONE;
	REFERENCE_TYPE *diff;
	REFERENCE_BASETYPE Tnrm, Rnrm, rerr;
	diff = (REFERENCE_TYPE *) malloc (size * m);
	mublasConvert2D (m, 1, 0, target, diff);

#if defined (PREC_D) || defined (PREC_S)
//	printf ("\n"); for (size_t i=0;i<m;i++) printf ("%zd: r:%1.3e, t:%1.3e\n", i, ref[i], target[i]);
#endif
// -----------------
#if defined (CHECK_CUBLAS)
	cublasHandle_t ch;
	cublasCreate (&ch);
	REFERENCE_TYPE *dev_t, *dev_r;
	cudaMalloc ((void **) &dev_t, size * m);
	cudaMalloc ((void **) &dev_r, size * m);
	cublasSetVector (m, size, ref, 1, dev_r, 1);
	cublasSetVector (m, size, diff, 1, dev_t, 1);
	REFERENCE_AXPY (ch, m, &negone, dev_r, 1, dev_t, 1);
	cublasGetVector (m, size, dev_t, 1, diff, 1);
	REFERENCE_NRM (ch, m, dev_r, 1, &Rnrm);
	REFERENCE_NRM (ch, m, dev_t, 1, &Tnrm);
	cudaFree (dev_r);
	cudaFree (dev_t);
#else
#if !defined (CHECK_MPACK) && ( defined (PREC_C) || defined (PREC_Z) || defined (PREC_U) || defined (PREC_V) )
	REFERENCE_AXPY (m, &negone, ref, 1, diff, 1);
#else
	REFERENCE_AXPY (m, negone, ref, 1, diff, 1);
#endif
	Rnrm = REFERENCE_NRM (m, ref, 1);
	Tnrm = REFERENCE_NRM (m, diff, 1);
#endif
// -----------------

	rerr = Tnrm / Rnrm;
#if defined (CHECK_CUBLAS)
	printf("\t%1.3e\t", rerr);
#elif defined (CHECK_CBLAS)
	printf("\t%1.3e\t", rerr);
#elif defined (CHECK_MPACK)
	printf("\t%1.3e\t(lo:\t%1.3e)\t", rerr.x[0], rerr.x[1]);
#endif
#if DEBUG_PRINT
	printf("\n%%--- result (debug-mode) ---\n");
#if defined (PREC_S) || defined (PREC_D)
	printf("%%index\tdiff\treference\ttarget\n");
	for (size_t i=0; i < m; i++)
		printf ("%zd\t%+1.2e\t%+1.2e\t%+1.2e\n", i, diff[i], ref[i], target[i]);
#endif
#if defined (PREC_C) || defined (PREC_Z)
	printf("%%index\tdiff(r,i)\treference(r,i)\ttarget(r,i)\n");
	for (size_t i=0; i < m; i++)
		#if defined (CHECK_CBLAS)
		printf ("%zd\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\n", (size_t)i, diff[i].r, diff[i].i, ref[i].r, ref[i].i, target[i].x, target[i].y);
		#elif defined (CHECK_CBLAS)
		printf ("%zd\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\t%+1.2e\n", (size_t)i, diff[i].x, diff[i].y, ref[i].x, ref[i].y, target[i].x, target[i].y);
		#endif
#endif
#endif
	free (diff);
}

// print routine -------------------------------------------------
void print_routine_name () {
	#if defined (PREC_S)
	printf ("S");
	#elif defined (PREC_D)
	printf ("D");
	#elif defined (PREC_C)
	printf ("C");
	#elif defined (PREC_Z)
	printf ("Z");
	#elif defined (PREC_W)
	printf ("W");
	#elif defined (PREC_U)
	printf ("U");
	#elif defined (PREC_B)
	printf ("B");
	#elif defined (PREC_V)
	printf ("V");
	#else
	printf ("unknown-precision,"
	#endif
	#if defined (GEMV)
	printf ("GEMV\n");
	#elif defined (SYMV)
	printf ("SYMV\n");
	#elif defined (HEMV)
	printf ("HEMV\n");
	#elif defined (TRMV)
	printf ("TRMV\n");
	#elif defined (TRSV)
	printf ("TRSV\n");
	#elif defined (SCAL)
	printf ("SCAL\n");
	#elif defined (AXPY)
	printf ("AXPY\n");
	#elif defined (COPY)
	printf ("COPY\n");
	#elif defined (DOT)
	printf ("DOT\n");
	#elif defined (DOTU)
	printf ("DOTU\n");
	#elif defined (GER)
	printf ("GER\n");
	#elif defined (CSRMV)
	printf ("CSRMV\n");
	#elif defined (HYBMV)
	printf ("HYBMV\n");
	#elif defined (GEMM)
	printf ("GEMM\n");
	#elif defined (GEAM)
	printf ("GEAM\n");
	#else
	printf ("unknown-routine\n"
	#endif
}

void print_library_name () {
	#if defined (MUBLAS)
	printf ("MUBLAS\n");
	#elif defined (CUBLAS)
	printf ("CUBLAS\n");
	#elif defined (KBLAS)
	printf ("KBLAS\n");
	#elif defined (MAGMA)
	printf ("MAGMA\n");
	#elif defined (CULA)
	printf ("CULA\n");
	#elif defined (ASPEN)
	printf ("ASPEN.K2\n");
	#else
	printf ("unknown-library\n");
	#endif
}

void print_info1 (testingHandle_t *th) {
	if (!th->nodisp) {
		system("echo -n '% Testing info -----------------------------\n'");
		system("echo -n '% \tDate\t'; date");
		system("echo -n '% \tHost\t'; hostname");
		system("echo -n '% \t'");
		system("head -n 1 /proc/driver/nvidia/version");
		system("echo -n '% \t'");
		system("tail -n 1 /proc/driver/nvidia/version");
		printf("%% \tNum of GPUs (installed)\t%d\n", th->num_installed_gpus);
		printf("%% \tNum of GPUs (used)\t%d\n", th->num_gpus);
		for (uint32_t igpu = th->devno; igpu < std::max (th->num_gpus, th->devno+1); igpu++) {
			printf("%% \tDevice[%d]\t%s\n", igpu, th->devprop[igpu].name);
			printf("%% \t- Theoretical Peak Memory Bandwidth\t%3.1f [GB/s]\n", th->device_bytes_per_sec[igpu] * 1.e-9);
			printf("%% \t- Theoretical Peak Performance (SP)\t%4.1f [GFlops]\n", th->device_flops_per_sec_sp[igpu] * 1.e-9);
			printf("%% \t- Theoretical Peak Performance (DP)\t%4.1f [GFlops]\n", th->device_flops_per_sec_dp[igpu] * 1.e-9);
			printf("%% \t- Total Gmem\t%4.1f [GB]\n", th->devprop[igpu].totalGlobalMem * 1e-9);
			printf("%% \t- CC\t%d.%d\n", th->devprop[igpu].major, th->devprop[igpu].minor);
			printf("%% \t- ECC Enabled\t%d\n", th->devprop[igpu].ECCEnabled);
			printf("%% \t- MultiProcessorCount\t%d\n", th->devprop[igpu].multiProcessorCount);
		}
		printf("%% \tDriver\t%d\n", th->driverVersion);
		printf("%% \tRuntime\t%d\n", th->runtimeVersion);
		printf ("%%\tRoutine\t");
		print_routine_name ();
		printf ("%%\tLibrary\t");
		print_library_name ();
		if (th->mode == 'c') {
			printf ("%%\tCheck\t");
			#if defined (CHECK_CUBLAS)
			printf ("CUBLAS\n");
			#elif defined (CHECK_CBLAS)
			printf ("CBLAS\n");
			#elif defined (CHECK_MPACK)
			printf ("MBLAS\n");
			#endif
		}
		printf("%% \tGmem allocation\t");
		#if PITCHALLOC
		printf("pitched\n");
		#else
		printf("non-pitched\n");
		#endif
		printf ("%%\tMINLOOP\t%d\n", MINLOOP);
		printf ("%%\tMAXLOOP\t%d\n", MAXLOOP);
		printf ("%%\tMINSEC\t%f\n", MINSEC);
		printf ("%% ------------------------------------------\n");
		printf ("%% BLAS parameter ---------------------------\n");
		#if !defined (DIM1) 
		printf ("%%\ttranA\t%c\n", th->tranA);
		#endif
		#if defined (DIM3) 
		printf ("%%\ttranB\t%c\n", th->tranB);
		#endif
		#if !defined (DIM1)
		printf ("%%\tuplo\t%c\n", th->uplo);
		printf ("%%\tdiag\t%c\n", th->diag);
		printf ("%%\tside\t%c\n", th->side);
		#endif
		printf ("%%\tdim_start\t%zd\n", (size_t)th->dim_start);
		printf ("%%\tdim_stop\t%zd\n", (size_t)th->dim_stop);
		printf ("%%\tdim_step\t%zd\n", (size_t)th->dim_step);
		#if !defined (DIM1)
		if (th->dim_m_const == 0) printf ("%%\tM\trange\n");
		else printf ("%%\tM\t%zd\n", (size_t)th->dim_m_const);
		#endif
		if (th->dim_n_const == 0) printf ("%%\tN\trange\n");
		else printf ("%%\tN\t%zd\n", (size_t)th->dim_n_const);
		#if defined (DIM3)
		if (th->dim_k_const == 0) printf ("%%\tK\trange\n");
		else printf ("%%\tK\t%zd\n", (size_t)th->dim_k_const);
		#endif
		#if !defined (DIM3) 
		printf ("%%\tincx\t%zd\n", (size_t)th->incx);
		#endif
		#if defined (DIM2) 
		printf ("%%\tincy\t%zd\n", (size_t)th->incy);
		#endif
		printf ("%%\tcublasatomic\t%d\n", th->cublasatomic);
		if (strcmp (th->mtx_file, "")) printf ("%%\tmtx_file\t%s\n", th->mtx_file);
		printf ("%% ------------------------------------------\n");
	}
}

void print_info2 (testingHandle_t *th) {
	if (!th->nodisp) {
		printf ("%% Evaluation result ------------------------\n");
		printf ("%% M\tN\tK");
		if (th->mode == 'p')
			printf ("\tTime[sec]\t[GFlops]\t[GB/s]\t(#iter)");
		if (th->mode == 'c')
			printf ("\tERROR");
		#if defined (MUBLAS)
		printf ("\tTx\tTy\tTh\tTyPerTx\tBx\tBy\tWrpOcp\tBlkOcp\tGrdOcp\tCount\tkernel#");
		#endif
		#if defined (USE_NVML)
		printf ("\tTemp.bf[C]\tTemp.af[C]");
		#endif
		printf ("\n");
	}
}

#if defined (MUBLAS)
void print_mublas_info1 (testingHandle_t *th, mublasHandle_t *mh) {
//	if (!th->nodisp) {
		printf ("\t%zd\t%zd\t%zd\t%1.3f\t%zd\t%zd\t%1.3f\t%1.3f\t%1.3f\t%d\t%d",
		(size_t)mh->Tx, (size_t)mh->Ty, (size_t)(mh->Tx*mh->Ty),
		(float)mh->Ty/(float)mh->Tx,
		(size_t)mh->Bx, (size_t)mh->By,
		(float)mh->WrpOcp, (float)mh->BlkOcp, (float)mh->GrdOcp,
		(int32_t)mh->count, (int32_t)mh->kernelnumber);
//	}
}

void print_mublas_info2 (testingHandle_t *th, mublasHandle_t *mh) {
	if (!th->nodisp) {
		printf ("%% ------------------------------------------\n");
		printf ("%% MUBLAS info (last called) ----------------\n");
		printf ("%% \tmublasVersion\t%d\n", mh->mublasVersion);
		printf ("%% \tbinaryVersion\t%d\n", mh->funcatb.binaryVersion);
		printf ("%% \tconstSizeBytes\t%zd\n", mh->funcatb.constSizeBytes);
		printf ("%% \tlocalSizeBytes\t%zd\n", mh->funcatb.localSizeBytes);
		printf ("%% \tmaxThreadsPerBlock\t%d\n", mh->funcatb.maxThreadsPerBlock);
		printf ("%% \tnumRegs\t%d\n", mh->funcatb.numRegs);
		printf ("%% \tptxVersion\t%d\n", mh->funcatb.ptxVersion);
		printf ("%% \tShmPerBlkMax\t%zd\n", mh->ShmPerBlkMax);
		printf ("%% ------------------------------------------\n");
		printf ("%% \tTxStatic\t%zd\n", (size_t)mh->TxStatic);
		printf ("%% \tTyStatic\t%zd\n", (size_t)mh->TyStatic);
		printf ("%% ------------------------------------------\n");
		printf ("%% \tThMin\t%zd\n", (size_t)mh->ThMin);
		printf ("%% \tThMax\t%zd\n", (size_t)mh->ThMax);
		printf ("%% \tThStep\t%zd\n", (size_t)mh->ThStep);
		printf ("%% \tTxMin\t%zd\n", (size_t)mh->TxMin);
		printf ("%% \tTxMax\t%zd\n", (size_t)mh->TxMax);
		printf ("%% \tTxStep\t%zd\n", (size_t)mh->TxStep);
		printf ("%% ------------------------------------------\n");
		printf ("%% \tWrpOcpSuf\t%1.4f\n", mh->WrpOcpSuf);
		printf ("%% \tBlkOcpSuf\t%1.4f\n", mh->BlkOcpSuf);
		printf ("%% \tGrdOcpSuf\t%1.4f\n", mh->GrdOcpSuf);
//		printf ("%% \tTxPrefer\t%c\n", mh->TxPrefer);
		printf ("%% \tTyPerTxMax\t%f\n", mh->TyPerTxMax);
		printf ("%% ------------------------------------------\n");
	}
}
#endif

void print_devprop (testingHandle_t *th, cudaDeviceProp *devprop) {
	if (!th->nodisp) {
		printf ("totalGlobalMem\t%zd\n", devprop->totalGlobalMem);
		printf ("sharedMemPerBlock\t%zd\n", devprop->sharedMemPerBlock);
		printf ("regsPerBlock\t%d\n", devprop->regsPerBlock);
		printf ("warpSize\t%d\n", devprop->warpSize);
		printf ("memPitch\t%zd\n", devprop->memPitch);
		printf ("maxThreadsPerBlock\t%d\n", devprop->maxThreadsPerBlock);
		printf ("maxThreadsDim[0]\t%d\n", devprop->maxThreadsDim[0]);
		printf ("maxThreadsDim[1]\t%d\n", devprop->maxThreadsDim[1]);
		printf ("maxThreadsDim[2]\t%d\n", devprop->maxThreadsDim[2]);
		printf ("maxGridSize[0]\t%d\n", devprop->maxGridSize[0]);
		printf ("maxGridSize[1]\t%d\n", devprop->maxGridSize[1]);
		printf ("maxGridSize[2]\t%d\n", devprop->maxGridSize[2]);
		printf ("clockRate\t%d\n", devprop->clockRate);
		printf ("totalConstMem\t%zd\n", devprop->totalConstMem);
		printf ("major\t%d\n", devprop->major);
		printf ("minor\t%d\n", devprop->minor);
		printf ("textureAlignment\t%zd\n", devprop->textureAlignment);
		printf ("texturePitrchAlignment\t%zd\n", devprop->texturePitchAlignment);
		printf ("deviceOverlap\t%d\n", devprop->deviceOverlap);
		printf ("multiProcessorCount\t%d\n",  devprop->multiProcessorCount);
		printf ("kernelExecTimeoutEnabled\t%d\n", devprop->kernelExecTimeoutEnabled);
		printf ("integrated\t%d\n", devprop->integrated);
		printf ("canMapHostMemory\t%d\n", devprop->canMapHostMemory);
		printf ("computeMode\t%d\n", devprop->computeMode);
		printf ("maxTexture1D\t%d\n", devprop->maxTexture1D);
		printf ("maxTexture1DMipmap\t%d\n", devprop->maxTexture1DMipmap);
		printf ("maxTexture1DLinear\t%d\n", devprop->maxTexture1DLinear);
		printf ("maxTexture2D[0]\t%d\n", devprop->maxTexture2D[0]);
		printf ("maxTexture2D[1]\t%d\n", devprop->maxTexture2D[1]);
		printf ("maxTexture2DMipmap[0]\t%d\n", devprop->maxTexture2DMipmap[0]);
		printf ("maxTexture2DMipmap[1]\t%d\n", devprop->maxTexture2DMipmap[1]);
		printf ("maxTexture2DLinear[0]\t%d\n", devprop->maxTexture2DLinear[0]);
		printf ("maxTexture2DLinear[1]\t%d\n", devprop->maxTexture2DLinear[1]);
		printf ("maxTexture2DLinear[2]\t%d\n", devprop->maxTexture2DLinear[2]);
		printf ("maxTexture2DGather[0]\t%d\n", devprop->maxTexture2DGather[0]);
		printf ("maxTexture2DGather[1]\t%d\n", devprop->maxTexture2DGather[1]);
		printf ("maxTexture3D[0]\t%d\n", devprop->maxTexture3D[0]);
		printf ("maxTexture3D[1]\t%d\n", devprop->maxTexture3D[1]);
		printf ("maxTexture3D[2]\t%d\n", devprop->maxTexture3D[2]);
		printf ("maxTexture3DAlt[0]\t%d\n", devprop->maxTexture3DAlt[0]);
		printf ("maxTexture3DAlt[1]\t%d\n", devprop->maxTexture3DAlt[1]);
		printf ("maxTexture3DAlt[2]\t%d\n", devprop->maxTexture3DAlt[2]);
		printf ("maxTextureCubemap\t%d\n", devprop->maxTextureCubemap);
		printf ("maxTexture1DLayered[0]\t%d\n", devprop->maxTexture1DLayered[0]);
		printf ("maxTexture1DLayered[1]\t%d\n", devprop->maxTexture1DLayered[1]);
		printf ("maxTexture2DLayered[0]\t%d\n", devprop->maxTexture2DLayered[0]);
		printf ("maxTexture2DLayered[1]\t%d\n", devprop->maxTexture2DLayered[1]);
		printf ("maxTexture2DLayered[2]\t%d\n", devprop->maxTexture2DLayered[2]);
		printf ("maxTextureCubemapLayered[0]\t%d\n", devprop->maxTextureCubemapLayered[0]);
		printf ("maxTextureCubemapLayered[1]\t%d\n", devprop->maxTextureCubemapLayered[1]);
		printf ("maxSurface1D\t%d\n", devprop->maxSurface1D);
		printf ("maxSurface2D[0]\t%d\n", devprop->maxSurface2D[0]);
		printf ("maxSurface2D[1]\t%d\n", devprop->maxSurface2D[1]);
		printf ("maxSurface3D[0]\t%d\n", devprop->maxSurface3D[0]);
		printf ("maxSurface3D[1]\t%d\n", devprop->maxSurface3D[1]);
		printf ("maxSurface3D[2]\t%d\n", devprop->maxSurface3D[2]);
		printf ("maxSurface1DLayered[0]\t%d\n", devprop->maxSurface1DLayered[0]);
		printf ("maxSurface1DLayered[1]\t%d\n", devprop->maxSurface1DLayered[1]);
		printf ("maxSurface2DLayered[0]\t%d\n", devprop->maxSurface2DLayered[0]);
		printf ("maxSurface2DLayered[1]\t%d\n", devprop->maxSurface2DLayered[1]);
		printf ("maxSurface2DLayered[2]\t%d\n", devprop->maxSurface2DLayered[2]);
		printf ("maxSurfaceCubemap\t%d\n", devprop->maxSurfaceCubemap);
		printf ("maxSurfaceCubemapLayered[0]\t%d\n", devprop->maxSurfaceCubemapLayered[0]);
		printf ("maxSurfaceCubemapLayered[1]\t%d\n", devprop->maxSurfaceCubemapLayered[1]);
		printf ("surfaceAlignment\t%zd\n", devprop->surfaceAlignment);
		printf ("concurrentKernels\t%d\n", devprop->concurrentKernels);
		printf ("ECCEnabled\t%d\n", devprop->ECCEnabled);
		printf ("pciBusID\t%d\n", devprop->pciBusID);
		printf ("pciDeviceID\t%d\n", devprop->pciDeviceID);
		printf ("pciDomainID\t%d\n", devprop->pciDomainID);
		printf ("tccDriver\t%d\n", devprop->tccDriver);
		printf ("asyncEngineCount\t%d\n", devprop->asyncEngineCount);
		printf ("unifiedAddressing\t%d\n", devprop->unifiedAddressing);
		printf ("memoryClockRate\t%d\n", devprop->memoryClockRate);
		printf ("memoryBusWidth\t%d\n", devprop->memoryBusWidth);
		printf ("l2CacheSize\t%d\n", devprop->l2CacheSize);
		printf ("maxThreadsPerMultiProcessor\t%d\n", devprop->maxThreadsPerMultiProcessor);
		printf ("streamPrioritiesSupported\t%d\n", devprop->streamPrioritiesSupported);
		printf ("globalL1CacheSupported\t%d\n", devprop->globalL1CacheSupported);
		printf ("localL1CacheSupported\t%d\n", devprop->localL1CacheSupported);
		printf ("sharedMemPerMultiprocessor\t%zd\n", devprop->sharedMemPerMultiprocessor);
		printf ("regsPerMultiprocessor\t%d\n", devprop->regsPerMultiprocessor);
//		printf ("managedMemSupported\t%d\n", devprop->managedMemSupported);
		printf ("isMultiGpuBoard\t%d\n", devprop->isMultiGpuBoard);
		printf ("multiGpuBoardGroupID\t%d\n", devprop->multiGpuBoardGroupID);
		printf ("----------------\n");
	}
}
