#include "testing_common.h"
#include "testing_level2.h"
#include "testing_common.cpp"

int32_t
main (int32_t argc, char **argv)
{
	size_t i, iter = MINLOOP;
	size_t lda_dev, lda_hst;
	size_t cda_dev, cda_hst;
	size_t rda_dev, rda_hst;
	size_t vlx_dev, vlx_hst;
	size_t vly_dev, vly_hst;
	size_t word_size_r = sizeof (REFERENCE_TYPE);
	double sec, t0, t1;
	cudaError_t err;

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	print_info1 (&th);
// --------------------------------------------

// library setup ------------------------------
	#if defined (CHECK_CUBLAS)
	cublasHandle_t ch;
	cublasCreate (&ch);
	cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
	#endif
	#if defined (MUBLAS)
	mublasHandle_t ha;
	mublasCreate (&ha);
	set_mublas_handler (&th, &ha);
	#elif defined (CUBLAS)
	cublasHandle_t ha;
	cublasCreate (&ha);
	cublasSetPointerMode(ha, CUBLAS_POINTER_MODE_HOST);
	if (th.cublasatomic == 1) 
		cublasSetAtomicsMode(ha, CUBLAS_ATOMICS_ALLOWED);
	else
		cublasSetAtomicsMode(ha, CUBLAS_ATOMICS_NOT_ALLOWED);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	magma_init();
	#elif defined (CULA)
    culaInitialize();
	#elif defined (ASPEN)
	ASPEN_init(0);
	#else
	#endif
#if DEBUG_ALLSKIP == 0
// --------------------------------------------

// memory setup -------------------------------
	GPU_TYPE alpha, beta;
	GPU_TYPE *hst_A, *hst_X, *hst_Y, *hst_Y_t;
	GPU_TYPE *dev_A, *dev_X, *dev_Y;
	REFERENCE_TYPE alpha_r = REFERENCE_ZERO, beta_r = REFERENCE_ZERO;
	REFERENCE_TYPE *hst_A_r = 0, *hst_X_r = 0, *hst_Y_r = 0;
	
	// dim setup
	dim_hst_setup (&th);
	rda_hst = th.dim_stop;
	cda_hst = th.dim_stop;
	vlx_hst = th.dim_stop;
	vly_hst = th.dim_stop;
	if (th.dim_m_const > 0) rda_hst = th.dim_m_const;
	if (th.dim_n_const > 0) cda_hst = th.dim_n_const;
	if (th.tranA == 'N' || th.tranA == 'n') {
		if (th.dim_m_const > 0) vly_hst = th.dim_m_const;
		if (th.dim_n_const > 0) vlx_hst = th.dim_n_const;
	} else {
		if (th.dim_n_const > 0) vly_hst = th.dim_n_const;
		if (th.dim_m_const > 0) vlx_hst = th.dim_m_const;
	}
	lda_hst = rda_hst;
	// malloc host memory
	cudaMallocHost ((void **) &hst_A, th.word_size * lda_hst * cda_hst);
	cudaMallocHost ((void **) &hst_X, th.word_size * vlx_hst * th.incx);
	cudaMallocHost ((void **) &hst_Y, th.word_size * vly_hst * th.incy);
	// initialize
	mublasInit2D (1, 1, 0, &alpha);
	mublasInit2D (1, 1, 0, &beta);
	mublasInit2D (lda_hst, cda_hst, lda_hst, hst_A);
	mublasInit2D (vlx_hst * th.incx, 1, 0, hst_X);
	mublasInit2D (vly_hst * th.incy, 1, 0, hst_Y);
	if (th.mode == 'c') {
		// initialize
		cudaMallocHost ((void **) &hst_A_r, word_size_r * lda_hst * cda_hst);
		cudaMallocHost ((void **) &hst_X_r, word_size_r * vlx_hst * th.incx);
		cudaMallocHost ((void **) &hst_Y_r, word_size_r * vly_hst * th.incy);
		cudaMallocHost ((void **) &hst_Y_t, th.word_size * vly_hst * th.incy);
	}
// --------------------------------------------

	print_info2 (&th);

// evaluation ---------------------------------
	dim_dev_setup (&th);
	while (1) {
		if ((th.dim_m_const == 0 && th.dim_m_dev > th.dim_stop) ||
		    (th.dim_n_const == 0 && th.dim_n_dev > th.dim_stop) ||
		    (th.dim_k_const == 0 && th.dim_k_dev > th.dim_stop)) break;
		// dim setup
		rda_dev = th.dim_m_dev;
		cda_dev = th.dim_n_dev;
		if (th.tranA == 'N' || th.tranA == 'n') {
			vlx_dev = th.dim_n_dev;
			vly_dev = th.dim_m_dev;
		} else {
			vlx_dev = th.dim_m_dev;
			vly_dev = th.dim_n_dev;
		}
		// malloc device memory
		#if PITCHALLOC
		size_t pitch;
		cudaMallocPitch ((void **) &dev_A, &pitch, th.word_size * rda_dev, cda_dev);
		lda_dev = pitch/th.word_size;
		#else
		lda_dev = rda_dev;
		cudaMalloc ((void **) &dev_A, th.word_size * lda_dev * cda_dev);
		#endif
		cudaMalloc ((void **) &dev_X, th.word_size * vlx_dev * th.incx);
		cudaMalloc ((void **) &dev_Y, th.word_size * vly_dev * th.incy);
		// memcpy from hst to device
		cublasSetMatrix (rda_dev, cda_dev, th.word_size, hst_A, lda_hst, dev_A, lda_dev);
		cublasSetVector (vlx_dev * th.incx, th.word_size, hst_X, 1, dev_X, 1);
		cublasSetVector (vly_dev * th.incy, th.word_size, hst_Y, 1, dev_Y, 1);
		// ---------------------------------------------

		if (th.mode != 't') {
			#if defined (DIM1)
			printf ("--\t%zd\t--", (size_t)th.dim_n_dev);
			#elif defined (DIM2)
			printf ("%zd\t%zd\t--", (size_t)th.dim_m_dev, (size_t)th.dim_n_dev);
			#endif
		}
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p') {
			#if PERF_ACCURATE
			iter = get_num_iter (&th);
			ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,alpha,dev_A,lda_dev,dev_X,beta,dev_Y); // dummy
			// +++++ measurement +++++++
			t0 = gettime ();
			for (i = 0; i < iter; i++) ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,alpha,dev_A,lda_dev,dev_X,beta,dev_Y);
			t1 = gettime ();
			// +++++++++++++++++++++++++
			sec = (t1 - t0) / (float)iter;
			#else
			float sec_all_exec[MINLOOP];
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				t0 = gettime ();
				ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,alpha,dev_A,lda_dev,dev_X,beta,dev_Y);
				t1 = gettime ();
				sec_all_exec[i] = (t1 - t0);
			}
			sec = 0.;
			for (i = 0; i < iter; i++) {
				if (sec_all_exec[i] > sec)
					sec = sec_all_exec[i];
			}
			#endif
			printf ("\t%1.3e\t%1.3e\t%1.3e\t(%zd)", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, (size_t)iter);
		}
		if (th.mode == 'c' || th.mode == 't') 
			ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,alpha,dev_A,lda_dev,dev_X,beta,dev_Y);
		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		// -------------------------------------------

		if (th.mode == 'c') 
			cublasGetVector (vly_dev * th.incy, th.word_size, dev_Y, 1, hst_Y_t, 1);

		cudaFree (dev_A);
		cudaFree (dev_X);
		cudaFree (dev_Y);

		if (th.mode == 'c') {
			mublasConvert2D (1, 1, 0, &alpha, &alpha_r);
			mublasConvert2D (1, 1, 0, &beta, &beta_r);
			mublasConvert2D (rda_dev, cda_dev, lda_hst, hst_A, hst_A_r);
			mublasConvert2D (vlx_dev * th.incx, 1, 0, hst_X, hst_X_r);
			mublasConvert2D (vly_dev * th.incy, 1, 0, hst_Y, hst_Y_r);
			#if (defined (CHECK_CUBLAS))
			REFERENCE_TYPE *dev_A_r, *dev_X_r, *dev_Y_r;
			cudaMalloc ((void **) &dev_A_r, word_size_r * lda_dev * cda_dev);
			cudaMalloc ((void **) &dev_X_r, word_size_r * vlx_dev * th.incx);
			cudaMalloc ((void **) &dev_Y_r, word_size_r * vly_dev * th.incy);
			cublasSetMatrix (rda_dev, cda_dev, word_size_r, hst_A_r, lda_hst, dev_A_r, lda_dev);
			cublasSetVector (vlx_dev * th.incx, word_size_r, hst_X_r, 1, dev_X_r, 1);
			cublasSetVector (vly_dev * th.incy, word_size_r, hst_Y_r, 1, dev_Y_r, 1);
			ROUTINE_REFERENCE(ch,th,th.dim_m_dev,th.dim_n_dev,alpha_r,dev_A_r,lda_dev,dev_X_r,beta_r,dev_Y_r);
			cublasGetVector (vly_dev, word_size_r, dev_Y_r, 1, hst_Y_r, 1);
			cudaFree (dev_A_r);
			cudaFree (dev_X_r);
			cudaFree (dev_Y_r);
			#else
			ROUTINE_REFERENCE(th,th.dim_m_dev,th.dim_n_dev,alpha_r,hst_A_r,lda_hst,hst_X_r,beta_r,hst_Y_r);
			#endif
			mublasCheckVector (vly_dev * th.incy, hst_Y_t, hst_Y_r);
		}

		if (th.mode != 't') {
			#if defined (MUBLAS)
			print_mublas_info1 (&th, &ha);
			#endif
		}
		printf ("\n");

		dim_dev_increment (&th);
	}
// --------------------------------------------

// shutdown -----------------------------------
	cudaFreeHost (hst_A);
	cudaFreeHost (hst_X);
	cudaFreeHost (hst_Y);
	if (th.mode == 'c') {
		cudaFreeHost (hst_A_r);
		cudaFreeHost (hst_X_r);
		cudaFreeHost (hst_Y_r);
		cudaFreeHost (hst_Y_t);
	}

	#if defined (CHECK_CUBLAS)
	cublasDestroy (ch);
	#endif
	#if defined (MUBLAS)
	print_mublas_info2 (&th, &ha);
	mublasDestroy (&ha);
	#elif defined (CUBLAS)
	cublasDestroy (ha);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	#elif defined (CULA)
	#elif defined (ASPEN)
	ASPEN_shutdown();
	#endif
// --------------------------------------------

#endif
	return 0;
}

