#include "testing_common.h"
#include "testing_level2.h"
#include "testing_common.cpp"

#include <kblas.h> 
#if defined (PREC_S)
#define KBLAS_GET_GEMV_MGPU_BS 	get_sgemv_mgpu_bs
#define KBLAS_MALLOC_MGPU_1D 	kblas_smalloc_mgpu_1D
#define KBLAS_SETMATRIX_MGPU_1D	kblas_ssetmatrix_mgpu_1D
#define KBLAS_GEMV_MGPU_ASYNC	kblas_sgemv_mgpu_async
#endif
#if defined (PREC_D)
#define KBLAS_GET_GEMV_MGPU_BS 	get_dgemv_mgpu_bs
#define KBLAS_MALLOC_MGPU_1D 	kblas_dmalloc_mgpu_1D
#define KBLAS_SETMATRIX_MGPU_1D	kblas_dsetmatrix_mgpu_1D
#define KBLAS_GEMV_MGPU_ASYNC	kblas_dgemv_mgpu_async
#endif
#if defined (PREC_C)
#define KBLAS_GET_GEMV_MGPU_BS 	get_cgemv_mgpu_bs
#define KBLAS_MALLOC_MGPU_1D 	kblas_cmalloc_mgpu_1D
#define KBLAS_SETMATRIX_MGPU_1D	kblas_csetmatrix_mgpu_1D
#define KBLAS_GEMV_MGPU_ASYNC	kblas_cgemv_mgpu_async
#endif
#if defined (PREC_Z)
#define KBLAS_GET_GEMV_MGPU_BS 	get_zgemv_mgpu_bs
#define KBLAS_MALLOC_MGPU_1D 	kblas_zmalloc_mgpu_1D
#define KBLAS_SETMATRIX_MGPU_1D	kblas_zsetmatrix_mgpu_1D
#define KBLAS_GEMV_MGPU_ASYNC	kblas_zgemv_mgpu_async
#endif

int32_t
main (int32_t argc, char **argv)
{
	size_t igpu;
	size_t i, iter = MINLOOP;
	size_t lda_hst;
	size_t cda_dev, cda_hst;
	size_t rda_dev, rda_hst;
	size_t vlx_dev, vlx_hst;
	size_t vly_dev, vly_hst;
	size_t word_size_r = sizeof (REFERENCE_TYPE);
	double sec, t0, t1;
	cudaError_t err;

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	print_info1 (&th);
// --------------------------------------------

// library setup ------------------------------
	#if defined (CHECK_CUBLAS)
	cublasHandle_t ch;
	cublasCreate (&ch);
	cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
	#endif
	cudaStream_t streams[NGPU_MAX][1];
	#if defined (MUBLAS)
	mublasHandle_t ha[NGPU_MAX];
	#endif
	for (igpu = 0; igpu < th.num_gpus; igpu++) {
		cudaSetDevice (igpu);
		cudaStreamCreate (&streams[igpu][0]);
		#if defined (MUBLAS)
		mublasCreate (&ha[igpu]);
		mublasSetStream (&ha[igpu], streams[igpu][0]);
		set_mublas_handler (&th, &ha[igpu]);
		#endif
	}
#if DEBUG_ALLSKIP == 0
// --------------------------------------------
	
// memory setup -------------------------------
	GPU_TYPE alpha, beta;
	GPU_TYPE *hst_A, *hst_X, *hst_Y, *hst_Y_t;
	REFERENCE_TYPE alpha_r = REFERENCE_ZERO, beta_r = REFERENCE_ZERO;
	REFERENCE_TYPE *hst_A_r = 0, *hst_X_r = 0, *hst_Y_r = 0;
	
	// dim setup
	dim_hst_setup (&th);
	rda_hst = th.dim_stop;
	cda_hst = th.dim_stop;
	vlx_hst = th.dim_stop;
	vly_hst = th.dim_stop;
	if (th.dim_m_const > 0) rda_hst = th.dim_m_const;
	if (th.dim_n_const > 0) cda_hst = th.dim_n_const;
	if (th.tranA == 'N' || th.tranA == 'n') {
		if (th.dim_m_const > 0) vly_hst = th.dim_m_const;
		if (th.dim_n_const > 0) vlx_hst = th.dim_n_const;
	} else {
		if (th.dim_n_const > 0) vly_hst = th.dim_n_const;
		if (th.dim_m_const > 0) vlx_hst = th.dim_m_const;
	}
	lda_hst = rda_hst;
	// malloc host memory
	cudaMallocHost ((void **) &hst_A, th.word_size * lda_hst * cda_hst);
	cudaMallocHost ((void **) &hst_X, th.word_size * vlx_hst * th.incx);
	cudaMallocHost ((void **) &hst_Y, th.word_size * vly_hst * th.incy);
	// initialize
	mublasInit2D (1, 1, 0, &alpha);
	mublasInit2D (1, 1, 0, &beta);
	mublasInit2D (lda_hst, cda_hst, lda_hst, hst_A);
	mublasInit2D (vlx_hst * th.incx, 1, 0, hst_X);
	mublasInit2D (vly_hst * th.incy, 1, 0, hst_Y);
	if (th.mode == 'c') {
		// initialize
		cudaMallocHost ((void **) &hst_A_r, word_size_r * lda_hst * cda_hst);
		cudaMallocHost ((void **) &hst_X_r, word_size_r * vlx_hst * th.incx);
		cudaMallocHost ((void **) &hst_Y_r, word_size_r * vly_hst * th.incy);
		cudaMallocHost ((void **) &hst_Y_t, th.word_size * vly_hst * th.incy);
	}
// --------------------------------------------

	print_info2 (&th);

// evaluation ---------------------------------
	dim_dev_setup (&th);
	while (1) {
		if ((th.dim_m_const == 0 && th.dim_m_dev > th.dim_stop) ||
		    (th.dim_n_const == 0 && th.dim_n_dev > th.dim_stop) ||
		    (th.dim_k_const == 0 && th.dim_k_dev > th.dim_stop)) break;

		if (th.tranA == 'N' || th.tranA == 'n') {
			vlx_dev = th.dim_n_dev;
			vly_dev = th.dim_m_dev;
		} else {
			vlx_dev = th.dim_m_dev;
			vly_dev = th.dim_n_dev;
		}
		rda_dev = th.dim_m_dev;
		cda_dev = th.dim_n_dev;
		GPU_TYPE* dev_A[NGPU_MAX] = {NULL};
		GPU_TYPE* dev_X[NGPU_MAX] = {NULL};
		GPU_TYPE* dev_Y[NGPU_MAX] = {NULL};
		GPU_TYPE* hst_Ysum[NGPU_MAX] = {NULL};
		size_t nb = KBLAS_GET_GEMV_MGPU_BS (th.tranA);
		size_t lda_dev = {((th.dim_m_dev+31)/32)*32};
		for (igpu = 0; igpu < th.num_gpus; igpu++) {
			cudaSetDevice (igpu);
			cudaMalloc((void**)&dev_X[igpu], vlx_dev * th.word_size);
			cudaMalloc((void**)&dev_Y[igpu], vly_dev * th.word_size);
			cudaMemcpy(dev_X[igpu], hst_X, vlx_dev * th.word_size, cudaMemcpyHostToDevice);
			cudaMemcpy(dev_Y[igpu], hst_Y, vly_dev * th.word_size, cudaMemcpyHostToDevice);
			hst_Ysum[igpu] = (GPU_TYPE*)malloc(vly_dev * th.word_size);
	   	}
		KBLAS_MALLOC_MGPU_1D (th.dim_m_dev, th.dim_n_dev, dev_A, th.num_gpus, lda_dev, nb);
		KBLAS_SETMATRIX_MGPU_1D (th.dim_m_dev, th.dim_n_dev, hst_A, lda_hst, dev_A, lda_dev, th.num_gpus, nb);

		#if defined (DIM1)
		printf ("--\t%zd\t--", th.dim_n_dev);
		#elif defined (DIM2)
		printf ("%zd\t%zd\t--", th.dim_m_dev, th.dim_n_dev);
		#endif
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p') {
			iter = get_num_iter (&th) * th.num_gpus;
			// +++++ measurement +++++++
			t0 = gettime ();
			for (i = 0; i < iter; i++) {
				#if defined (KBLAS)
				KBLAS_GEMV_MGPU_ASYNC (th.tranA, th.dim_m_dev, th.dim_n_dev, alpha, dev_A, lda_dev, dev_X, th.incx, beta, dev_Y, th.incy, th.num_gpus, 0, 0, streams);
				#endif
				#if defined (MUBLAS)
				MUBLAS_GEMV_MGPU_ASYNC (ha, th.tranA, th.dim_m_dev, th.dim_n_dev, alpha, dev_A, lda_dev, dev_X, th.incx, beta, dev_Y, th.incy, th.num_gpus, nb);
				#endif
			}
			for (igpu = 0; igpu < th.num_gpus; igpu++) {
				cudaSetDevice (igpu);
				cudaStreamSynchronize (streams[igpu][0]);
			}
			t1 = gettime ();
			// +++++++++++++++++++++++++
			sec = (t1 - t0) / (float)iter;
			printf ("\t%1.3e\t%1.3e\t%1.3e\t(%zd)", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, (size_t)iter);
		}
		if (th.mode == 'c') {
			#if defined (KBLAS)
			KBLAS_GEMV_MGPU_ASYNC (th.tranA, th.dim_m_dev, th.dim_n_dev, alpha, dev_A, lda_dev, dev_X, th.incx, beta, dev_Y, th.incy, th.num_gpus, 0, 0, streams);
			#endif
			#if defined (MUBLAS)
			MUBLAS_GEMV_MGPU_ASYNC (ha, th.tranA, th.dim_m_dev, th.dim_n_dev, alpha, dev_A, lda_dev, dev_X, th.incx, beta, dev_Y, th.incy, th.num_gpus, nb);
			#endif
		}
		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		// -------------------------------------------

		if (th.mode == 'c') {
			for (igpu = 0; igpu < th.num_gpus; igpu++) {
				cudaSetDevice (igpu);
				cudaMemcpy (hst_Ysum[igpu], dev_Y[igpu], vly_dev * th.word_size, cudaMemcpyDeviceToHost);
			}
			for(i = 0; i < vly_dev; i++) {
				//printf ("[igpu=0] %f, ", hst_Ysum[0][i]);
				for (igpu = 1; igpu < th.num_gpus; igpu++) {
					//printf ("[igpu=%d] %f", igpu, hst_Ysum[igpu][i]);
					#if defined (PREC_S) || defined (PREC_D)
     					hst_Ysum[0][i] += hst_Ysum[igpu][i];
					#else
	      				hst_Ysum[0][i].x += hst_Ysum[igpu][i].x;
      					hst_Ysum[0][i].y += hst_Ysum[igpu][i].y;
					#endif
				}
				//printf ("\n");
				hst_Y_t[i] = hst_Ysum[0][i];
			}
		}

		for (igpu = 0; igpu < th.num_gpus; igpu++) {
			cudaSetDevice (igpu);
			cudaFree (dev_A[igpu]);
			cudaFree (dev_X[igpu]);
			cudaFree (dev_Y[igpu]);
			free (hst_Ysum[igpu]);
		}

		if (th.mode == 'c') {
			mublasConvert2D (1, 1, 0, &alpha, &alpha_r);
			mublasConvert2D (1, 1, 0, &beta, &beta_r);
			mublasConvert2D (rda_dev, cda_dev, lda_hst, hst_A, hst_A_r);
			mublasConvert2D (vlx_dev * th.incx, 1, 0, hst_X, hst_X_r);
			mublasConvert2D (vly_dev * th.incy, 1, 0, hst_Y, hst_Y_r);
			#if (defined (CHECK_CUBLAS))
			REFERENCE_TYPE *dev_A_r, *dev_X_r, *dev_Y_r;
			cudaMalloc ((void **) &dev_A_r, word_size_r * rda_dev * cda_dev);
			cudaMalloc ((void **) &dev_X_r, word_size_r * vlx_dev * th.incx);
			cudaMalloc ((void **) &dev_Y_r, word_size_r * vly_dev * th.incy);
			cublasSetMatrix (rda_dev, cda_dev, word_size_r, hst_A_r, lda_hst, dev_A_r, rda_dev);
			cublasSetVector (vlx_dev * th.incx, word_size_r, hst_X_r, 1, dev_X_r, 1);
			cublasSetVector (vly_dev * th.incy, word_size_r, hst_Y_r, 1, dev_Y_r, 1);
			ROUTINE_REFERENCE(ch,th,th.dim_m_dev,th.dim_n_dev,alpha_r,dev_A_r,rda_dev,dev_X_r,beta_r,dev_Y_r);
			cublasGetVector (vly_dev, word_size_r, dev_Y_r, 1, hst_Y_r, 1);
			cudaFree (dev_A_r);
			cudaFree (dev_X_r);
			cudaFree (dev_Y_r);
			#else
			ROUTINE_REFERENCE(th,th.dim_m_dev,th.dim_n_dev,alpha_r,hst_A_r,lda_hst,hst_X_r,beta_r,hst_Y_r);
			#endif
			mublasCheckVector (vly_dev * th.incy, hst_Y_t, hst_Y_r);
		}

		#if defined (MUBLAS)
		print_mublas_info1 (&th, &ha[0]);
		#endif
		printf ("\n");

		dim_dev_increment (&th);
	}
// --------------------------------------------

// shutdown -----------------------------------
	cudaFreeHost (hst_A);
	cudaFreeHost (hst_X);
	cudaFreeHost (hst_Y);
	if (th.mode == 'c') {
		cudaFreeHost (hst_A_r);
		cudaFreeHost (hst_X_r);
		cudaFreeHost (hst_Y_r);
		cudaFreeHost (hst_Y_t);
	}

	#if defined (CHECK_CUBLAS)
	cublasDestroy (ch);
	#endif
	for (igpu = 0; igpu < th.num_gpus; igpu++) {
		#if defined (MUBLAS)
		printf ("%% (devno=%zd)\n", igpu);
		print_mublas_info2 (&th, &ha[igpu]);
		mublasDestroy (&ha[igpu]);
		#endif
		cudaStreamDestroy (streams[igpu][0]);
	}
// --------------------------------------------

#endif
	return 0;
}

